"""Agora SDP manipulation logic mimicking agoraRTC_N.js."""

import logging
from typing import Any

_LOGGER = logging.getLogger(__name__)


class SDPParser:
    """Basic SDP parser to avoid external dependencies, matching Agora JS behavior."""

    @staticmethod
    def parse(sdp: str) -> dict[str, Any]:
        parsed = {"media": []}
        current_media = None

        for line in sdp.splitlines():
            line = line.strip()
            if not line:
                continue

            parts = line.split("=", 1)
            if len(parts) < 2:
                continue
            ltype, lval = parts

            if ltype == "v":
                parsed["version"] = lval
            elif ltype == "o":
                oparts = lval.split()
                if len(oparts) >= 6:
                    parsed["origin"] = {
                        "username": oparts[0],
                        "sessionId": oparts[1],
                        "sessionVersion": oparts[2],
                        "netType": oparts[3],
                        "ipVer": oparts[4],
                        "address": oparts[5],
                    }
            elif ltype == "s":
                parsed["name"] = lval
            elif ltype == "m":
                mparts = lval.split()
                current_media = {
                    "type": mparts[0],
                    "port": int(mparts[1]),
                    "protocol": mparts[2],
                    "payloads": " ".join(mparts[3:]),
                    "rtp": [],
                    "fmtp": [],
                    "rtcpFb": [],
                    "ext": [],
                    "fingerprints": [],
                    "attributes": {},
                }
                parsed["media"].append(current_media)
            elif ltype == "a":
                aparts = lval.split(":", 1)
                attr = aparts[0]
                val = aparts[1] if len(aparts) > 1 else None

                target = current_media if current_media else parsed

                if attr == "ice-ufrag":
                    target["iceUfrag"] = val
                elif attr == "ice-pwd":
                    target["icePwd"] = val
                elif attr == "fingerprint":
                    fparts = val.split()
                    fp_obj = {"hash": fparts[0], "fingerprint": fparts[1]}
                    target["fingerprints"] = target.get("fingerprints", [])
                    target["fingerprints"].append(fp_obj)
                    # Keep backward compatibility for 'fingerprint' key if needed by other code?
                    # But parse_offer_to_ortc should use fingerprints list.
                    target["fingerprint"] = fp_obj
                elif attr == "setup":
                    target["setup"] = val
                elif attr == "mid":
                    target["mid"] = val
                elif attr == "direction":
                    target["direction"] = val
                elif attr == "ice-options":
                    target["iceOptions"] = val
                elif attr == "rtpmap":
                    rparts = val.split(None, 1)
                    pt = int(rparts[0])
                    rmap = rparts[1].split("/")
                    target["rtp"].append(
                        {
                            "payload": pt,
                            "codec": rmap[0],
                            "rate": int(rmap[1]) if len(rmap) > 1 else 90000,
                            "encoding": rmap[2] if len(rmap) > 2 else None,
                        }
                    )
                elif attr == "fmtp":
                    fparts = val.split(None, 1)
                    target["fmtp"].append({"payload": int(fparts[0]), "config": fparts[1]})
                elif attr == "rtcp-fb":
                    fbparts = val.split()
                    target["rtcpFb"].append(
                        {
                            "payload": int(fbparts[0]),
                            "type": fbparts[1],
                            "subtype": " ".join(fbparts[2:]) if len(fbparts) > 2 else None,
                        }
                    )
                elif attr == "extmap":
                    eparts = val.split()
                    target["ext"].append({"value": int(eparts[0]), "uri": eparts[1]})
                elif attr == "group":
                    if "groups" not in parsed:
                        parsed["groups"] = []
                    gparts = val.split()
                    parsed["groups"].append({"type": gparts[0], "mids": " ".join(gparts[1:])})
                elif attr == "msid-semantic":
                    parsed["msidSemantic"] = {
                        "semantic": val.split()[0],
                        "token": val.split()[1] if len(val.split()) > 1 else "",
                    }
        return parsed

    @staticmethod
    def write(parsed: dict[str, Any]) -> str:
        lines = [f"v={parsed.get('version', 0)}"]
        orig = parsed.get("origin", {})
        lines.append(
            f"o={orig.get('username', '-')} {orig.get('sessionId', 0)} {orig.get('sessionVersion', 0)} {orig.get('netType', 'IN')} IP{orig.get('ipVer', 4)} {orig.get('address', '127.0.0.1')}"
        )
        lines.append(f"s={parsed.get('name', '-')}")
        lines.append("t=0 0")
        for g in parsed.get("groups", []):
            lines.append(f"a=group:{g['type']} {g['mids']}")
        if "msidSemantic" in parsed:
            lines.append(f"a=msid-semantic: {parsed['msidSemantic']['semantic']} {parsed['msidSemantic']['token']}")
        if "icelite" in parsed:
            lines.append("a=ice-lite")
        if "extmapAllowMixed" in parsed:
            lines.append("a=extmap-allow-mixed")

        for m in parsed.get("media", []):
            lines.append(f"m={m['type']} {m['port']} {m['protocol']} {m['payloads']}")
            lines.append(f"c=IN IP4 {m.get('connection', {}).get('ip', '0.0.0.0')}")
            if "rtcp" in m:
                lines.append(
                    f"a=rtcp:{m['rtcp']['port']} IN IP4 {m['rtcp'].address if hasattr(m['rtcp'], 'address') else m['rtcp'].get('address', '0.0.0.0')}"
                )
            if "iceUfrag" in m:
                lines.append(f"a=ice-ufrag:{m['iceUfrag']}")
            if "icePwd" in m:
                lines.append(f"a=ice-pwd:{m['icePwd']}")
            if "iceOptions" in m:
                lines.append(f"a=ice-options:{m['iceOptions']}")
            if "fingerprint" in m and not m.get("fingerprints"):
                lines.append(f"a=fingerprint:{m['fingerprint']['hash']} {m['fingerprint']['fingerprint']}")
            for fp in m.get("fingerprints", []):
                lines.append(f"a=fingerprint:{fp['hash']} {fp['fingerprint']}")
            if "setup" in m:
                lines.append(f"a=setup:{m['setup']}")
            if "mid" in m:
                lines.append(f"a=mid:{m['mid']}")
            if "direction" in m:
                lines.append(f"a={m['direction']}")
            for r in m.get("rtp", []):
                val = f"{r['payload']} {r['codec']}/{r['rate']}"
                if r.get("encoding"):
                    val += f"/{r['encoding']}"
                lines.append(f"a=rtpmap:{val}")
            for fb in m.get("rtcpFb", []):
                val = f"{fb['payload']} {fb['type']}"
                if fb.get("subtype"):
                    val += f" {fb['subtype']}"
                lines.append(f"a=rtcp-fb:{val}")
            for f in m.get("fmtp", []):
                lines.append(f"a=fmtp:{f['payload']} {f['config']}")
            for e in m.get("ext", []):
                lines.append(f"a=extmap:{e['value']} {e['uri']}")
            if "rtcpMux" in m:
                lines.append("a=rtcp-mux")
            if "rtcpRsize" in m:
                lines.append("a=rtcp-rsize")
            for s in m.get("ssrcs", []):
                lines.append(f"a=ssrc:{s['id']} {s['attribute']}:{s['value']}")
            for c in m.get("candidates", []):
                val = f"{c['foundation']} {c['component']} {c['protocol']} {c['priority']} {c['ip']} {c['port']} typ {c['type']}"
                lines.append(f"a=candidate:{val}")
        return "\r\n".join(lines) + "\r\n"


def parse_offer_to_ortc(offer_sdp: str) -> dict[str, Any]:
    """Parse offer SDP to ORTC object, matching Agora SDK getOrtc logic."""
    parsed = SDPParser.parse(offer_sdp)
    ice_params = {}
    dtls_params = {}

    # Helper: Check if codec can be sent
    def can_send(codec_obj: dict) -> bool:
        name = codec_obj["rtpMap"]["encodingName"].upper()
        params = codec_obj.get("fmtp", {}).get("parameters", {})

        if name == "H265":
            return False
        if name == "VP9":
            # Profile 0 and 2 are sendable, 1 and 3 are not
            pid = params.get("profile-id")
            if pid in ("1", "3"):
                return False
        if name == "AV1":
            # Profile 1 is recv only
            if params.get("profile") == "1":
                return False
        return True

    # Global/Session level parameters
    if "iceUfrag" in parsed:
        ice_params = {"iceUfrag": parsed["iceUfrag"], "icePwd": parsed["icePwd"]}
    if parsed.get("fingerprints"):
        dtls_params = {
            "fingerprints": [
                {"hashFunction": fp["hash"], "fingerprint": fp["fingerprint"]} for fp in parsed["fingerprints"]
            ]
        }

    # Iterate media sections to extract params and build caps
    caps = {
        "send": {
            "audioCodecs": [],
            "audioExtensions": [],
            "videoCodecs": [],
            "videoExtensions": [],
        },
        "recv": {
            "audioCodecs": [],
            "audioExtensions": [],
            "videoCodecs": [],
            "videoExtensions": [],
        },
        "sendrecv": {
            "audioCodecs": [],
            "audioExtensions": [],
            "videoCodecs": [],
            "videoExtensions": [],
        },
    }

    for m in parsed.get("media", []):
        # Extract ICE/DTLS from first media section if not found globally
        if not ice_params and "iceUfrag" in m:
            ice_params = {"iceUfrag": m["iceUfrag"], "icePwd": m["icePwd"]}
        if not dtls_params and m.get("fingerprints"):
            dtls_params = {
                "fingerprints": [
                    {"hashFunction": fp["hash"], "fingerprint": fp["fingerprint"]} for fp in m["fingerprints"]
                ]
            }

        mtype = m.get("type")
        codecs = []

        # Parse codecs
        for rtp in m.get("rtp", []):
            pt = rtp.get("payload")
            codec = {
                "payloadType": pt,
                "rtpMap": {
                    "encodingName": rtp.get("codec"),
                    "clockRate": rtp.get("rate"),
                },
                "rtcpFeedbacks": [],
                "fmtp": {"parameters": {}},
            }
            if rtp.get("encoding"):
                codec["rtpMap"]["encodingParameters"] = int(rtp.get("encoding"))

            # Feedbacks
            for fb in m.get("rtcpFb", []):
                if fb.get("payload") == pt:
                    fb_obj = {"type": fb.get("type")}
                    if fb.get("subtype"):
                        fb_obj["parameter"] = fb.get("subtype")
                    codec["rtcpFeedbacks"].append(fb_obj)

            # Add forced rrtr if missing
            if not any(fb["type"] == "rrtr" for fb in codec["rtcpFeedbacks"]):
                codec["rtcpFeedbacks"].append({"type": "rrtr"})

            # FMTP
            for f in m.get("fmtp", []):
                if f.get("payload") == pt:
                    for part in f.get("config", "").split(";"):
                        if "=" in part:
                            k, v = part.split("=", 1)
                            codec["fmtp"]["parameters"][k.strip()] = v.strip()
                        elif part.strip():
                            # Handle flags or key-only params if any (less common in fmtp but possible)
                            # JS logic: params[k.trim()] = v ? v.trim() : null;
                            codec["fmtp"]["parameters"][part.strip()] = None
            codecs.append(codec)

        # Parse extensions
        extensions = [{"entry": ext.get("value"), "extensionName": ext.get("uri")} for ext in m.get("ext", [])]

        # Distribute codecs based on can_send logic
        for codec in codecs:
            is_send = can_send(codec)
            is_recv = True  # Assumption: browser can recv what it offers

            target_lists = []
            if is_send and is_recv:
                target_lists.append(caps["sendrecv"])
            elif is_recv:
                target_lists.append(caps["recv"])

            for t in target_lists:
                if mtype == "audio":
                    t["audioCodecs"].append(codec)
                elif mtype == "video":
                    t["videoCodecs"].append(codec)

        # Extensions usually go to sendrecv
        target_ext = caps["sendrecv"]
        if mtype == "audio":
            target_ext["audioExtensions"].extend(extensions)
        elif mtype == "video":
            target_ext["videoExtensions"].extend(extensions)

    return {
        "iceParameters": ice_params,
        "dtlsParameters": dtls_params,
        "rtpCapabilities": caps,
        "version": "2",
    }


def generate_answer_from_ortc(
    ortc_params: dict[str, Any],
    offer_sdp: dict[str, Any],
    force_setup: str | None = None,
) -> str:
    """Generate answer SDP from ORTC parameters (from join_success)."""
    # ortc_params is from join_success (Agora)
    # offer_sdp is the dict returned by parse_offer_to_ortc(offer_sdp)
    dtls = ortc_params.get("dtlsParameters", {})
    ice = ortc_params.get("iceParameters", {})
    rtp_caps = ortc_params.get("rtpCapabilities", {})
    cname = ortc_params.get("cname", "")
    offer_parsed = offer_sdp

    # setup logic from yx(): server -> passive, client -> active, auto -> actpass
    role = dtls.get("role", "server")
    setup = force_setup or ("passive" if role == "server" else "active" if role == "client" else "actpass")

    answer = {
        "version": 0,
        "origin": {
            "username": "-",
            "sessionId": 0,
            "sessionVersion": 0,
            "netType": "IN",
            "ipVer": 4,
            "address": "127.0.0.1",
        },
        "name": "AgoraGateway",
        "groups": offer_parsed.get("groups", [{"type": "BUNDLE", "mids": "0 1"}]),
        "msidSemantic": offer_parsed.get("msidSemantic", {"semantic": "WMS", "token": ""}),
        "icelite": "ice-lite",
        "media": [],
        "extmapAllowMixed": "extmap-allow-mixed",
    }

    # Match IDs from original offer
    offer_ext_map = {
        ext.get("extensionName"): ext.get("entry") for m in offer_parsed.get("media", []) for ext in m.get("ext", [])
    }

    for idx, offer_m in enumerate(offer_parsed.get("media", [])):
        mtype = offer_m.get("type", "audio")
        mid = offer_m.get("mid", str(idx))

        # RTP capabilities can be flat or nested
        caps = rtp_caps
        if "recv" in rtp_caps:
            caps = rtp_caps["recv"]
        elif "sendrecv" in rtp_caps:
            caps = rtp_caps["sendrecv"]

        codecs = caps.get("videoCodecs" if mtype == "video" else "audioCodecs", [])
        extensions = caps.get("videoExtensions" if mtype == "video" else "audioExtensions", [])

        answer_m = {
            "type": mtype,
            "port": 9,
            "protocol": "UDP/TLS/RTP/SAVPF",
            "payloads": " ".join([str(c.get("payloadType")) for c in codecs]),
            "connection": {"version": 4, "ip": "0.0.0.0"},
            "rtcp": {"port": 9, "netType": "IN", "ipVer": 4, "address": "0.0.0.0"},
            "iceUfrag": ice.get("iceUfrag"),
            "icePwd": ice.get("icePwd"),
            "iceOptions": "trickle",
            "fingerprints": [],
            "setup": setup,
            "mid": mid,
            "direction": "sendonly",
            "rtp": [],
            "rtcpFb": [],
            "fmtp": [],
            "ext": [],
            "rtcpMux": "rtcp-mux",
            "rtcpRsize": "rtcp-rsize",
        }

        # Handle multiple fingerprints
        dtls_fps = dtls.get("fingerprints", [])
        if not dtls_fps:
            # Fallback if single
            dtls_fps = [{"hashFunction": "sha-256", "fingerprint": ""}]  # Should usually be present

        for fp in dtls_fps:
            answer_m["fingerprints"].append(
                {
                    "hash": fp.get("hashFunction", fp.get("algorithm", "sha-256")),
                    "fingerprint": fp.get("fingerprint"),
                }
            )

        for c in codecs:
            pt = c.get("payloadType")
            answer_m["rtp"].append(
                {
                    "payload": pt,
                    "codec": c["rtpMap"].get("encodingName"),
                    "rate": c["rtpMap"].get("clockRate"),
                    "encoding": c["rtpMap"].get("encodingParameters"),
                }
            )
            for fb in c.get("rtcpFeedbacks", []):
                answer_m["rtcpFb"].append(
                    {
                        "payload": pt,
                        "type": fb.get("type"),
                        "subtype": fb.get("parameter"),
                    }
                )

            # handle fmtp
            fmtp_params = c.get("fmtp", {}).get("parameters", {})

            # Match JS: Force stereo for opus
            if c["rtpMap"].get("encodingName", "").lower() == "opus":
                fmtp_params["stereo"] = "1"
                fmtp_params["sprop-stereo"] = "1"

            if fmtp_params:
                config_parts = []
                for k, v in fmtp_params.items():
                    if v is not None:
                        config_parts.append(f"{k}={v}")
                    else:
                        config_parts.append(k)

                answer_m["fmtp"].append(
                    {
                        "payload": pt,
                        "config": ";".join(config_parts),
                    }
                )

        for ext in extensions:
            uri = ext.get("extensionName")
            if uri in offer_ext_map:
                answer_m["ext"].append({"value": offer_ext_map[uri], "uri": uri})

        # Note: Working SDK for audience role does NOT include SSRC lines in the answer
        # if cname: answer_m["ssrcs"] = [{"id": 12345678 if mtype == "audio" else 87654321, "attribute": "cname", "value": cname}]

        # Add Candidates from iceParameters.candidates if any
        if "candidates" in ice:
            answer_m["candidates"] = []
            for c in ice["candidates"]:
                answer_m["candidates"].append(
                    {
                        "foundation": c.get("foundation", "0"),
                        "component": c.get("component", 1),
                        "protocol": c.get("protocol", "udp"),
                        "priority": c.get("priority", 0),
                        "ip": c.get("ip"),
                        "port": c.get("port"),
                        "type": c.get("type"),
                    }
                )

        answer["media"].append(answer_m)

    return SDPParser.write(answer)



"""Aliyun IoT MQTT client for PyMammotion."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from enum import Enum, auto
import hashlib
import hmac
import json
from logging import getLogger
import ssl
import threading
from typing import TYPE_CHECKING

import paho.mqtt.client as mqtt
from paho.mqtt.enums import CallbackAPIVersion

if TYPE_CHECKING:
    from paho.mqtt.reasoncodes import ReasonCode

from pymammotion.aliyun.cloud_gateway import CloudIOTGateway

logger = getLogger(__name__)

# Aliyun IoT broker CA certificate bundle (GlobalSign + Aliyun IoT roots)
_ALIYUN_BROKER_CA_DATA = """
-----BEGIN CERTIFICATE-----
MIIDdTCCAl2gAwIBAgILBAAAAAABFUtaw5QwDQYJKoZIhvcNAQEFBQAwVzELMAkG
A1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExEDAOBgNVBAsTB1Jv
b3QgQ0ExGzAZBgNVBAMTEkdsb2JhbFNpZ24gUm9vdCBDQTAeFw05ODA5MDExMjAw
MDBaFw0yODAxMjgxMjAwMDBaMFcxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9i
YWxTaWduIG52LXNhMRAwDgYDVQQLEwdSb290IENBMRswGQYDVQQDExJHbG9iYWxT
aWduIFJvb3QgQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDaDuaZ
jc6j40+Kfvvxi4Mla+pIH/EqsLmVEQS98GPR4mdmzxzdzxtIK+6NiY6arymAZavp
xy0Sy6scTHAHoT0KMM0VjU/43dSMUBUc71DuxC73/OlS8pF94G3VNTCOXkNz8kHp
1Wrjsok6Vjk4bwY8iGlbKk3Fp1S4bInMm/k8yuX9ifUSPJJ4ltbcdG6TRGHRjcdG
snUOhugZitVtbNV4FpWi6cgKOOvyJBNPc1STE4U6G7weNLWLBYy5d4ux2x8gkasJ
U26Qzns3dLlwR5EiUWMWea6xrkEmCMgZK9FGqkjWZCrXgzT/LCrBbBlDSgeF59N8
9iFo7+ryUp9/k5DPAgMBAAGjQjBAMA4GA1UdDwEB/wQEAwIBBjAPBgNVHRMBAf8E
BTADAQH/MB0GA1UdDgQWBBRge2YaRQ2XyolQL30EzTSo//z9SzANBgkqhkiG9w0B
AQUFAAOCAQEA1nPnfE920I2/7LqivjTFKDK1fPxsnCwrvQmeU79rXqoRSLblCKOz
yj1hTdNGCbM+w6DjY1Ub8rrvrTnhQ7k4o+YviiY776BQVvnGCv04zcQLcFGUl5gE
38NflNUVyRRBnMRddWQVDf9VMOyGj/8N7yy5Y0b2qvzfvGn9LhJIZJrglfCm7ymP
AbEVtQwdpf5pLGkkeB6zpxxxYu7KyJesF12KwvhHhm4qxFYxldBniYUr+WymXUad
DKqC5JlR3XC321Y9YeRq4VzW9v493kHMB65jUr9TU/Qr6cf9tveCX4XSQRjbgbME
HMUfpIBvFSDJ3gyICh3WZlXi/EjJKSZp4A==
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIID3zCCAsegAwIBAgISfiX6mTa5RMUTGSC3rQhnestIMA0GCSqGSIb3DQEBCwUA
MHcxCzAJBgNVBAYTAkNOMREwDwYDVQQIDAhaaGVqaWFuZzERMA8GA1UEBwwISGFu
Z3pob3UxEzARBgNVBAoMCkFsaXl1biBJb1QxEDAOBgNVBAsMB1Jvb3QgQ0ExGzAZ
BgNVBAMMEkFsaXl1biBJb1QgUm9vdCBDQTAgFw0yMzA3MDQwNjM2NThaGA8yMDUz
MDcwNDA2MzY1OFowdzELMAkGA1UEBhMCQ04xETAPBgNVBAgMCFpoZWppYW5nMREw
DwYDVQQHDAhIYW5nemhvdTETMBEGA1UECgwKQWxpeXVuIElvVDEQMA4GA1UECwwH
Um9vdCBDQTEbMBkGA1UEAwwSQWxpeXVuIElvVCBSb290IENBMIIBIjANBgkqhkiG
9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoK//6vc2oXhnvJD7BVhj6grj7PMlN2N4iNH4
GBmLmMdkF1z9eQLjksYc4Zid/FX67ypWFtdycOei5ec0X00m53Gvy4zLGBo2uKgi
T9IxMudmt95bORZbaph4VK82gPNU4ewbiI1q2loRZEHRdyPORTPpvNLHu8DrYBnY
Vg5feEYLLyhxg5M1UTrT/30RggHpaa0BYIPxwsKyylQ1OskOsyZQeOyPe8t8r2D4
RBpUGc5ix4j537HYTKSyK3Hv57R7w1NzKtXoOioDOm+YySsz9sTLFajZkUcQci4X
aedyEeguDLAIUKiYicJhRCZWljVlZActorTgjCY4zRajodThrQIDAQABo2MwYTAO
BgNVHQ8BAf8EBAMCAQYwDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUkWHoKi2h
DlS1/rYpcT/Ue+aKhP8wHwYDVR0jBBgwFoAUkWHoKi2hDlS1/rYpcT/Ue+aKhP8w
DQYJKoZIhvcNAQELBQADggEBADrrLcBY7gDXN8/0KHvPbGwMrEAJcnF9z4MBxRvt
rEoRxhlvRZzPi7w/868xbipwwnksZsn0QNIiAZ6XzbwvIFG01ONJET+OzDy6ZqUb
YmJI09EOe9/Hst8Fac2D14Oyw0+6KTqZW7WWrP2TAgv8/Uox2S05pCWNfJpRZxOv
Lr4DZmnXBJCMNMY/X7xpcjylq+uCj118PBobfH9Oo+iAJ4YyjOLmX3bflKIn1Oat
vdJBtXCj3phpfuf56VwKxoxEVR818GqPAHnz9oVvye4sQqBp/2ynrKFxZKUaJtk0
7UeVbtecwnQTrlcpWM7ACQC0OO0M9+uNjpKIbksv1s11xu0=
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIIFgzCCA2ugAwIBAgIORea7A4Mzw4VlSOb/RVEwDQYJKoZIhvcNAQEMBQAwTDE
gMB4GA1UECxMXR2xvYmFsU2lnbiBSb290IENBIC0gUjYxEzARBgNVBAoTCkdsb2
JhbFNpZ24xEzARBgNVBAMTCkdsb2JhbFNpZ24wHhcNMTQxMjEwMDAwMDAwWhcNM
zQxMjEwMDAwMDAwWjBMMSAwHgYDVQQLExdHbG9iYWxTaWduIFJvb3QgQ0EgLSBS
NjETMBEGA1UEChMKR2xvYmFsU2lnbjETMBEGA1UEAxMKR2xvYmFsU2lnbjCCAiI
wDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAJUH6HPKZvnsFMp7PPcNCPG0RQ
ssgrRIxutbPK6DuEGSMxSkb3/pKszGsIhrxbaJ0cay/xTOURQh7ErdG1rG1ofuT
ToVBu1kZguSgMpE3nOUTvOniX9PeGMIyBJQbUJmL025eShNUhqKGoC3GYEOfsSK
vGRMIRxDaNc9PIrFsmbVkJq3MQbFvuJtMgamHvm566qjuL++gmNQ0PAYid/kD3n
16qIfKtJwLnvnvJO7bVPiSHyMEAc4/2ayd2F+4OqMPKq0pPbzlUoSB239jLKJz9
CgYXfIWHSw1CM69106yqLbnQneXUQtkPGBzVeS+n68UARjNN9rkxi+azayOeSsJ
Da38O+2HBNXk7besvjihbdzorg1qkXy4J02oW9UivFyVm4uiMVRQkQVlO6jxTiW
m05OWgtH8wY2SXcwvHE35absIQh1/OZhFj931dmRl4QKbNQCTXTAFO39OfuD8l4
UoQSwC+n+7o/hbguyCLNhZglqsQY6ZZZZwPA1/cnaKI0aEYdwgQqomnUdnjqGBQ
Ce24DWJfncBZ4nWUx2OVvq+aWh2IMP0f/fMBH5hc8zSPXKbWQULHpYT9NLCEnFl
WQaYw55PfWzjMpYrZxCRXluDocZXFSxZba/jJvcE+kNb7gu3GduyYsRtYQUigAZ
cIN5kZeR1BonvzceMgfYFGM8KEyvAgMBAAGjYzBhMA4GA1UdDwEB/wQEAwIBBjA
PBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBSubAWjkxPioufi1xzWx/B/yGdToD
AfBgNVHSMEGDAWgBSubAWjkxPioufi1xzWx/B/yGdToDANBgkqhkiG9w0BAQwFA
AOCAgEAgyXt6NH9lVLNnsAEoJFp5lzQhN7craJP6Ed41mWYqVuoPId8AorRbrcW
c+ZfwFSY1XS+wc3iEZGtIxg93eFyRJa0lV7Ae46ZeBZDE1ZXs6KzO7V33EByrKP
rmzU+sQghoefEQzd5Mr6155wsTLxDKZmOMNOsIeDjHfrYBzN2VAAiKrlNIC5waN
rlU/yDXNOd8v9EDERm8tLjvUYAGm0CuiVdjaExUd1URhxN25mW7xocBFymFe944
Hn+Xds+qkxV/ZoVqW/hpvvfcDDpw+5CRu3CkwWJ+n1jez/QcYF8AOiYrg54NMMl
+68KnyBr3TsTjxKM4kEaSHpzoHdpx7Zcf4LIHv5YGygrqGytXm3ABdJ7t+uA/iU
3/gKbaKxCXcPu9czc8FB10jZpnOZ7BN9uBmm23goJSFmH63sUYHpkqmlD75HHTO
wY3WzvUy2MmeFe8nI+z1TIvWfspA9MRf/TuTAjB0yPEL+GltmZWrSZVxykzLsVi
VO6LAUP5MSeGbEYNNVMnbrt9x+vJJUEeKgDu+6B5dpffItKoZB0JaezPkvILFa9
x8jvOOJckvB595yEunQtYQEgfn7R8k8HWV+LLUNS60YMlOH1Zkd5d9VUWx+tJDf
LRVpOoERIyNiwmcUVhAn21klJwGW45hpxbqCo8YLoRT5s1gLXCmeDBVrJpBA=
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIICCzCCAZGgAwIBAgISEdK7ujNu1LzmJGjFDYQdmOhDMAoGCCqGSM49BAMDME
YxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9iYWxTaWduIG52LXNhMRwwGgYD
VQQDExNHbG9iYWxTaWduIFJvb3QgRTQ2MB4XDTE5MDMyMDAwMDAwMFoXDTQ2MD
MyMDAwMDAwMFowRjELMAkGA1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24g
bnYtc2ExHDAaBgNVBAMTE0dsb2JhbFNpZ24gUm9vdCBFNDYwdjAQBgcqhkjOPQ
IBBgUrgQQAIgNiAAScDrHPt+ieUnd1NPqlRqetMhkytAepJ8qUuwzSChDH2omw
lwxwEwkBjtjqR+q+soArzfwoDdusvKSGN+1wCAB16pMLey5SnCNoIwZD7JIvU4
Tb+0cUB+hflGddyXqBPCCjQjBAMA4GA1UdDwEB/wQEAwIBhjAPBgNVHRMBAf8E
BTADAQH/MB0GA1UdDgQWBBQxCpCPtsad0kRLgLWi5h+xEk8blTAKBggqhkjOPQ
QDAwNoADBlAjEA31SQ7Zvvi5QCkxeCmb6zniz2C5GMn0oUsfZkvLtoURMMA/cV
i4RguYv/Uo7njLwcAjA8+RHUjE7AwWHCFUyqqx0LMV87HOIAl0Qx5v5zli/alt
P+CAezNIm8BZ/3Hobui3A=
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIIFWjCCA0KgAwIBAgISEdK7udcjGJ5AXwqdLdDfJWfRMA0GCSqGSIb3DQEBDAU
AMEYxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9iYWxTaWduIG52LXNhMRwwGg
YDVQQDExNHbG9iYWxTaWduIFJvb3QgUjQ2MB4XDTE5MDMyMDAwMDAwMFoXDTQ2M
DMyMDAwMDAwMFowRjELMAkGA1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24g
bnYtc2ExHDAaBgNVBAMTE0dsb2JhbFNpZ24gUm9vdCBSNDYwggIiMA0GCSqGSIb
3DQEBAQUAA4ICDwAwggIKAoICAQCsrHQy6LNl5brtQyYdpokNRbopiLKkHWPd08
EsCVeJOaFV6Wc0dwxu5FUdUiXSE2te4R2pt32JMl8Nnp8semNgQB+msLZ4j5lUl
ghYruQGvGIFAha/r6gjA7aUD7xubMLL1aa7DOn2wQL7Id5m3RerdELv8HQvJfTq
a1VbkNud316HCkD7rRlr+/fKYIje2sGP1q7Vf9Q8g+7XFkyDRTNrJ9CG0Bwta/O
rffGFqfUo0q3v84RLHIf8E6M6cqJaESvWJ3En7YEtbWaBkoe0G1h6zD8K+kZPT
Xhc+CtI4wSEy132tGqzZfxCnlEmIyDLPRT5ge1lFgBPGmSXZgjPjHvjK8Cd+RTy
G/FWaha/LIWFzXg4mutCagI0GIMXTpRW+LaCtfOW3T3zvn8gdz57GSNrLNRyc0N
XfeD412lPFzYE+cCQYDdF3uYM2HSNrpyibXRdQr4G9dlkbgIQrImwTDsHTUB+JM
WKmIJ5jqSngiCNI/onccnfxkF0oE32kRbcRoxfKWMxWXEM2G/CtjJ9++ZdU6Z+F
fy7dXxd7Pj2Fxzsx2sZy/N78CsHpdlseVR2bJ0cpm4O6XkMqCNqo98bMDGfsVR7
/mrLZqrcZdCinkqaByFrgY/bxFn63iLABJzjqls2k+g9vXqhnQt2sQvHnf3PmKg
Gwvgqo6GDoLclcqUC4wIDAQABo0IwQDAOBgNVHQ8BAf8EBAMCAYYwDwYDVR0TAQ
H/BAUwAwEB/zAdBgNVHQ4EFgQUA1yrc4GHqMywptWU4jaWSf8FmSwwDQYJKoZIh
vcNAQEMBQADggIBAHx47PYCLLtbfpIrXTncvtgdokIzTfnvpCo7RGkerNlFo048
p9gkUbJUHJNOxO97k4VgJuoJSOD1u8fpaNK7ajFxzHmuEajwmf3lH7wvqMxX63b
EIaZHU1VNaL8FpO7XJqti2kM3S+LGteWygxk6x9PbTZ4IevPuzz5i+6zoYMzRx6
Fcg0XERczzF2sUyQQCPtIkpnnpHs6i58FZFZ8d4kuaPp92CC1r2LpXFNqD6v6MV
enQTqnMdzGxRBF6XLE+0xRFFRhiJBPSy03OXIPBNvIQtQ6IbbjhVp+J3pZmOUdk
LG5NrmJ7v2B0GbhWrJKsFjLtrWhV/pi60zTe9Mlhww6G9kuEYO4Ne7UyWHmRVSy
BQ7N0H3qqJZ4d16GLuc1CLgSkZoNNiTW2bKg2SnkheCLQQrzRQDGQob4Ez8pn7f
XwgNNgyYMqIgXQBztSvwyeqiv5u+YfjyW6hY0XHgL+XVAEV8/+LbzvXMAaq7afJ
Mbfc2hIkCwU9D9SGuTSyxTDYWnP4vkYxboznxSjBF25cfe1lNj2M8FawTSLfJvd
kzrnE6JwYZ+vj+vYxXX4M2bUdGc6N3ec592kD3ZDZopD8p/7DEJ4Y9HiD2971KE
9dJeFt0g5QdYg/NA6s/rob8SKunE3vouXsXgxT7PntgMTzlSdriVZzH81Xwj3QE
UxeCp6
-----END CERTIFICATE-----
"""

_MQTT_PORT = 8883
_MQTT_KEEPALIVE = 60
_MQTT_MAX_INFLIGHT = 20
_MQTT_MAX_QUEUED = 40
_MQTT_RECONNECT_MIN_SEC = 1
_MQTT_RECONNECT_MAX_SEC = 60


class _State(Enum):
    DISCONNECTED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    DISCONNECTING = auto()


class AliyunMQTT:
    """Async-friendly MQTT client for the Aliyun IoT platform.

    Handles the Aliyun-specific credential format and topic structure while
    using paho-mqtt v2 directly (no LinkKit wrapper). Auto-reconnects with
    exponential backoff via paho's built-in reconnect_delay_set.
    """

    def __init__(
        self,
        region_id: str,
        product_key: str,
        device_name: str,
        device_secret: str,
        iot_token: str,
        cloud_client: CloudIOTGateway,
        client_id: str | None = None,
    ) -> None:
        super().__init__()
        self._cloud_client = cloud_client
        self._product_key = product_key
        self._device_name = device_name

        self.is_connected = False
        self.is_ready = False
        self.on_connected: Callable[[], Awaitable[None]] | None = None
        self.on_ready: Callable[[], Awaitable[None]] | None = None
        self.on_error: Callable[[str], Awaitable[None]] | None = None
        self.on_disconnected: Callable[[], Awaitable[None]] | None = None
        self.on_message: Callable[[str, str, str], Awaitable[None]] | None = None

        # Aliyun-specific credential format.
        # Client ID: "{base_id}|securemode=2,signmethod=hmacsha1|"
        if client_id is None:
            client_id = f"python-{device_name}"
        self._mqtt_client_id = f"{client_id}|securemode=2,signmethod=hmacsha1|"

        # Username: "{device_name}&{product_key}"
        self._mqtt_username = f"{device_name}&{product_key}"

        # Password: HMAC-SHA1("{device_secret}", "clientId{id}deviceName{dn}productKey{pk}")
        sign_content = f"clientId{client_id}deviceName{device_name}productKey{product_key}"
        self._mqtt_password = hmac.new(
            device_secret.encode("utf-8"), sign_content.encode("utf-8"), hashlib.sha1
        ).hexdigest()

        # Aliyun regional MQTT endpoint
        self._mqtt_host = f"{product_key}.iot-as-mqtt.{region_id}.aliyuncs.com"

        self._state = _State.DISCONNECTED
        self._state_lock = threading.Lock()
        self.loop = asyncio.get_running_loop()

        self._mqtt_client = self._build_mqtt_client()

    def _build_mqtt_client(self) -> mqtt.Client:
        """Build and configure the paho MQTT client with Aliyun TLS and settings."""
        client = mqtt.Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self._mqtt_client_id,
            protocol=mqtt.MQTTv311,
            transport="tcp",
            clean_session=True,
        )

        # TLS using Aliyun's CA bundle (GlobalSign + Aliyun IoT root certificates)
        tls_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cadata=_ALIYUN_BROKER_CA_DATA)
        client.tls_set_context(tls_context)

        client.username_pw_set(self._mqtt_username, self._mqtt_password)

        # Exponential backoff: starts at 1 s, doubles each attempt, caps at 60 s
        client.reconnect_delay_set(_MQTT_RECONNECT_MIN_SEC, _MQTT_RECONNECT_MAX_SEC)
        client.max_inflight_messages_set(_MQTT_MAX_INFLIGHT)
        client.max_queued_messages_set(_MQTT_MAX_QUEUED)

        client.on_connect = self._on_connect
        client.on_disconnect = self._on_disconnect
        client.on_message = self._on_message

        return client

    @property
    def iot_token(self) -> str:
        if authcode_response := self._cloud_client.session_by_authcode_response.data:
            return authcode_response.iotToken
        return ""

    def connect_async(self) -> None:
        """Start a non-blocking connection to the Aliyun MQTT broker.

        Safe to call from any state — calls while already connecting or connected
        are silently ignored.
        """
        with self._state_lock:
            if self._state in (_State.CONNECTED, _State.CONNECTING):
                logger.debug("Already %s, ignoring connect_async()", self._state.name)
                return
            self._state = _State.CONNECTING

        logger.info("Connecting to %s:%d", self._mqtt_host, _MQTT_PORT)
        self._mqtt_client.connect_async(self._mqtt_host, port=_MQTT_PORT, keepalive=_MQTT_KEEPALIVE)
        self._mqtt_client.loop_start()

    def disconnect(self) -> None:
        """Disconnect from the Aliyun MQTT broker and stop the network loop."""
        with self._state_lock:
            if self._state == _State.DISCONNECTED:
                return
            self._state = _State.DISCONNECTING

        logger.info("Disconnecting from Aliyun MQTT")
        self._mqtt_client.disconnect()
        self._mqtt_client.loop_stop()

    # ------------------------------------------------------------------
    # paho callbacks (called from paho's background thread)
    # ------------------------------------------------------------------

    def _on_connect(
        self,
        client: mqtt.Client,
        userdata: None,
        connect_flags: mqtt.ConnectFlags,
        reason_code: ReasonCode,
        properties: mqtt.Properties | None,
    ) -> None:
        if reason_code.is_failure:
            logger.error("Connection refused by broker: %s", reason_code)
            with self._state_lock:
                self._state = _State.DISCONNECTED
            return

        logger.info("Connected to Aliyun MQTT broker")
        with self._state_lock:
            self._state = _State.CONNECTED
        self.is_connected = True

        # Subscribe to all Aliyun device-specific down-topics
        for topic in self._subscription_topics():
            client.subscribe(topic, qos=1)

        # Bind this client session to the device using its IoT token
        client.publish(
            f"/sys/{self._product_key}/{self._device_name}/app/up/account/bind",
            json.dumps(
                {
                    "id": "msgid1",
                    "version": "1.0",
                    "request": {"clientId": self._mqtt_username},
                    "params": {"iotToken": self.iot_token},
                }
            ),
            qos=1,
        )

        if self.on_connected is not None:
            asyncio.run_coroutine_threadsafe(self.on_connected(), self.loop)

        self.is_ready = True
        if self.on_ready is not None:
            asyncio.run_coroutine_threadsafe(self.on_ready(), self.loop)

    def _on_disconnect(
        self,
        client: mqtt.Client,
        userdata: None,
        disconnect_flags: mqtt.DisconnectFlags,
        reason_code: ReasonCode,
        properties: mqtt.Properties | None,
    ) -> None:
        with self._state_lock:
            intentional = self._state == _State.DISCONNECTING
            self._state = _State.DISCONNECTED

        self.is_connected = False
        self.is_ready = False

        if intentional:
            logger.info("Disconnected (intentional)")
        else:
            logger.warning("Disconnected unexpectedly (rc=%s) — paho will reconnect", reason_code)

        if self.on_disconnected is not None:
            asyncio.run_coroutine_threadsafe(self.on_disconnected(), self.loop)

    def _on_message(self, client: mqtt.Client, userdata: None, message: mqtt.MQTTMessage) -> None:
        """Route incoming MQTT messages to the registered async handler."""
        logger.debug("Message on topic %s", message.topic)
        try:
            payload = json.loads(message.payload)
            logger.debug("Topic data %s", payload)
        except (json.JSONDecodeError, ValueError):
            logger.warning("Non-JSON payload on topic %s, ignoring", message.topic)
            return

        iot_id: str = payload.get("params", {}).get("iotId", "")
        if iot_id and self.on_message is not None:
            asyncio.run_coroutine_threadsafe(
                self.on_message(message.topic, message.payload.decode(), iot_id),
                self.loop,
            )

    # ------------------------------------------------------------------
    # Topic helpers
    # ------------------------------------------------------------------

    def _subscription_topics(self) -> list[str]:
        """Return the full list of Aliyun down-topics to subscribe to on connect."""
        base = f"/sys/{self._product_key}/{self._device_name}"
        return [
            f"{base}/app/down/account/bind_reply",
            f"{base}/app/down/thing/event/property/post_reply",
            f"{base}/app/down/thing/wifi/status/notify",
            f"{base}/app/down/thing/wifi/connect/event/notify",
            f"{base}/app/down/_thing/event/notify",
            f"{base}/app/down/thing/events",
            f"{base}/app/down/thing/status",
            f"{base}/app/down/thing/properties",
            f"{base}/app/down/thing/model/down_raw",
        ]

    def unsubscribe(self) -> None:
        """Unsubscribe from all device down-topics."""
        for topic in self._subscription_topics():
            self._mqtt_client.unsubscribe(topic)

    async def send_cloud_command(self, iot_id: str, command: bytes) -> str:
        """Send a command via the Aliyun cloud gateway."""
        return await self._cloud_client.send_cloud_command(iot_id, command)



from pymammotion.proto import basestation_pb2 as _basestation_pb2
from pymammotion.proto import mctrl_driver_pb2 as _mctrl_driver_pb2
from pymammotion.proto import mctrl_nav_pb2 as _mctrl_nav_pb2
from pymammotion.proto import mctrl_sys_pb2 as _mctrl_sys_pb2
from pymammotion.proto import dev_net_pb2 as _dev_net_pb2
from pymammotion.proto import mctrl_ota_pb2 as _mctrl_ota_pb2
from pymammotion.proto import luba_mul_pb2 as _luba_mul_pb2
from pymammotion.proto import mctrl_pept_pb2 as _mctrl_pept_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
DEV_BASESTATION: MsgDevice
DEV_BMS: MsgDevice
DEV_COMM_ESP: MsgDevice
DEV_IOTCTRL: MsgDevice
DEV_IOTSERVER: MsgDevice
DEV_LEFTMOTOR: MsgDevice
DEV_LOCALIZATION: MsgDevice
DEV_MAINCTL: MsgDevice
DEV_MOBILEAPP: MsgDevice
DEV_NAVIGATION: MsgDevice
DEV_PERCEPTION: MsgDevice
DEV_RIGHTMOTOR: MsgDevice
DEV_RTKCLI: MsgDevice
DEV_USBHOST: MsgDevice
MSG_ATTR_NONE: MsgAttr
MSG_ATTR_REPORT: MsgAttr
MSG_ATTR_REQ: MsgAttr
MSG_ATTR_RESP: MsgAttr
MSG_CMD_TYPE_APPLICATION: MsgCmdType
MSG_CMD_TYPE_BASESTATION: MsgCmdType
MSG_CMD_TYPE_EMBED_DRIVER: MsgCmdType
MSG_CMD_TYPE_EMBED_MIDWARE: MsgCmdType
MSG_CMD_TYPE_EMBED_OTA: MsgCmdType
MSG_CMD_TYPE_EMBED_SYS: MsgCmdType
MSG_CMD_TYPE_ESP: MsgCmdType
MSG_CMD_TYPE_LOCALIZATION: MsgCmdType
MSG_CMD_TYPE_MUL: MsgCmdType
MSG_CMD_TYPE_NAV: MsgCmdType
MSG_CMD_TYPE_PEPT: MsgCmdType
MSG_CMD_TYPE_PLANNING: MsgCmdType
MSG_CMD_TYPE_START: MsgCmdType
SOC_MODULE_MULTIMEDIA: MsgDevice

class LubaMsg(_message.Message):
    __slots__ = ["base", "driver", "msgattr", "msgtype", "mul", "nav", "net", "null", "ota", "pept", "rcver", "sender", "seqs", "subtype", "sys", "timestamp", "version"]
    BASE_FIELD_NUMBER: _ClassVar[int]
    DRIVER_FIELD_NUMBER: _ClassVar[int]
    MSGATTR_FIELD_NUMBER: _ClassVar[int]
    MSGTYPE_FIELD_NUMBER: _ClassVar[int]
    MUL_FIELD_NUMBER: _ClassVar[int]
    NAV_FIELD_NUMBER: _ClassVar[int]
    NET_FIELD_NUMBER: _ClassVar[int]
    NULL_FIELD_NUMBER: _ClassVar[int]
    OTA_FIELD_NUMBER: _ClassVar[int]
    PEPT_FIELD_NUMBER: _ClassVar[int]
    RCVER_FIELD_NUMBER: _ClassVar[int]
    SENDER_FIELD_NUMBER: _ClassVar[int]
    SEQS_FIELD_NUMBER: _ClassVar[int]
    SUBTYPE_FIELD_NUMBER: _ClassVar[int]
    SYS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    base: _basestation_pb2.BaseStation
    driver: _mctrl_driver_pb2.MctlDriver
    msgattr: MsgAttr
    msgtype: MsgCmdType
    mul: _luba_mul_pb2.SocMul
    nav: _mctrl_nav_pb2.MctlNav
    net: _dev_net_pb2.DevNet
    null: MsgNull
    ota: _mctrl_ota_pb2.MctlOta
    pept: _mctrl_pept_pb2.MctlPept
    rcver: MsgDevice
    sender: MsgDevice
    seqs: int
    subtype: int
    sys: _mctrl_sys_pb2.MctlSys
    timestamp: int
    version: int
    def __init__(self, msgtype: _Optional[_Union[MsgCmdType, str]] = ..., sender: _Optional[_Union[MsgDevice, str]] = ..., rcver: _Optional[_Union[MsgDevice, str]] = ..., msgattr: _Optional[_Union[MsgAttr, str]] = ..., seqs: _Optional[int] = ..., version: _Optional[int] = ..., subtype: _Optional[int] = ..., net: _Optional[_Union[_dev_net_pb2.DevNet, _Mapping]] = ..., sys: _Optional[_Union[_mctrl_sys_pb2.MctlSys, _Mapping]] = ..., nav: _Optional[_Union[_mctrl_nav_pb2.MctlNav, _Mapping]] = ..., driver: _Optional[_Union[_mctrl_driver_pb2.MctlDriver, _Mapping]] = ..., ota: _Optional[_Union[_mctrl_ota_pb2.MctlOta, _Mapping]] = ..., mul: _Optional[_Union[_luba_mul_pb2.SocMul, _Mapping]] = ..., null: _Optional[_Union[MsgNull, _Mapping]] = ..., pept: _Optional[_Union[_mctrl_pept_pb2.MctlPept, _Mapping]] = ..., base: _Optional[_Union[_basestation_pb2.BaseStation, _Mapping]] = ..., timestamp: _Optional[int] = ...) -> None: ...

class MsgNull(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class MsgCmdType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MsgAttr(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MsgDevice(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



"""RTK device information."""
from dataclasses import dataclass, field

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class RTK(DataClassORJSONMixin):
    """RTK device information."""

    device_id: str = field(metadata=field_options(alias="deviceId"))
    device_name: str = field(metadata=field_options(alias="deviceName"))
    product_key: str = field(metadata=field_options(alias="productKey"))
    status: int
    lora: str



from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pymammotion.proto import NavGetCommDataAck, NavGetHashListAck, SvgMessageAckT
from pymammotion.utility.mur_mur_hash import MurMurHashUtil


class PathType(IntEnum):
    """Path types for common data."""

    AREA = 0
    OBSTACLE = 1
    PATH = 2
    LINE = 10
    DUMP = 12
    SVG = 13
    VISUAL_SAFETY_ZONE = 25


@dataclass
class CommDataCouple:
    x: float = 0.0
    y: float = 0.0


@dataclass
class AreaLabelName(DataClassORJSONMixin):
    label: str = ""


@dataclass
class NavNameTime(DataClassORJSONMixin):
    name: str = ""
    create_time: int = 0
    modify_time: int = 0


@dataclass
class NavGetCommData(DataClassORJSONMixin):
    pver: int = 0
    sub_cmd: int = 0
    result: int = 0
    action: int = 0
    type: int = 0
    hash: int = 0
    paternal_hash_a: int = 0
    paternal_hash_b: int = 0
    total_frame: int = 0
    current_frame: int = 0
    data_hash: int = 0
    data_len: int = 0
    data_couple: list["CommDataCouple"] = field(default_factory=list)
    reserved: str = ""
    name_time: NavNameTime = field(default_factory=NavNameTime)


@dataclass
class MowPathPacket(DataClassORJSONMixin):
    path_hash: int = 0
    path_type: int = 0
    path_total: int = 0
    path_cur: int = 0
    zone_hash: int = 0
    data_couple: list["CommDataCouple"] = field(default_factory=list)


@dataclass
class MowPath(DataClassORJSONMixin):
    pver: int = 0
    sub_cmd: int = 0
    result: int = 0
    area: int = 0
    time: int = 0
    total_frame: int = 0
    current_frame: int = 0
    total_path_num: int = 0
    valid_path_num: int = 0
    data_hash: int = 0
    transaction_id: int = 0
    reserved: list[int] = field(default_factory=list)
    data_len: int = 0
    path_packets: list[MowPathPacket] = field(default_factory=list)


@dataclass
class SvgMessageData(DataClassORJSONMixin):
    x_move: float = 0.0
    y_move: float = 0.0
    scale: float = 0.0
    rotate: float = 0.0
    base_width_m: float = 0.0
    base_width_pix: int = 0
    base_height_m: float = 0.0
    base_height_pix: int = 0
    name_count: int = 0
    data_count: int = 0
    hide_svg: bool = False
    svg_file_name: str = ""
    svg_file_data: str = ""


@dataclass
class SvgMessage(DataClassORJSONMixin):
    pver: int = 0
    sub_cmd: int = 0
    total_frame: int = 0
    current_frame: int = 0
    data_hash: int = 0
    paternal_hash_a: int = 0
    type: int = 0
    result: int = 0
    svg_message: "SvgMessageData" = field(default_factory=SvgMessageData)


@dataclass
class FrameList(DataClassORJSONMixin):
    total_frame: int = 0
    sub_cmd: int = 0
    data: list[NavGetCommData | SvgMessage] = field(default_factory=list)


@dataclass
class Plan(DataClassORJSONMixin):
    pver: int = 0
    sub_cmd: int = 2
    area: int = 0
    work_time: int = 0
    version: str = ""
    id: str = ""
    user_id: str = ""
    device_id: str = ""
    plan_id: str = ""
    task_id: str = ""
    job_id: str = ""
    start_time: str = ""
    end_time: str = ""
    week: int = 0
    knife_height: int = 0
    model: int = 0
    edge_mode: int = 0
    required_time: int = 0
    route_angle: int = 0
    route_model: int = 0
    route_spacing: int = 0
    ultrasonic_barrier: int = 0
    total_plan_num: int = 0
    plan_index: int = 0
    result: int = 0
    speed: float = 0.0
    task_name: str = ""
    job_name: str = ""
    zone_hashs: list[int] = field(default_factory=list)
    reserved: str = ""
    start_date: str = ""
    end_date: str = ""
    trigger_type: int = 0
    day: int = 0
    weeks: list[int] = field(default_factory=list)
    remained_seconds: int = 0
    toward_mode: int = 0
    toward_included_angle: int = 0


@dataclass(eq=False, repr=False)
class NavGetHashListData(DataClassORJSONMixin):
    """Dataclass for NavGetHashListData."""

    pver: int = 0
    sub_cmd: int = 0
    total_frame: int = 0
    current_frame: int = 0
    data_hash: int = 0
    hash_len: int = 0
    reserved: str = ""
    result: int = 0
    data_couple: list[int] = field(default_factory=list)


@dataclass
class RootHashList(DataClassORJSONMixin):
    total_frame: int = 0
    sub_cmd: int = 0
    data: list[NavGetHashListData] = field(default_factory=list)


@dataclass
class AreaHashNameList(DataClassORJSONMixin):
    """Wrapper so we can serialize to and from dict."""

    name: str
    hash: int


@dataclass
class HashList(DataClassORJSONMixin):
    """stores our map data.
    [hashID, FrameList].
    hashlist for all our hashIDs for verification
    """

    root_hash_lists: list[RootHashList] = field(default_factory=list)
    area: dict[int, FrameList] = field(default_factory=dict)  # type 0
    path: dict[int, FrameList] = field(default_factory=dict)  # type 2
    obstacle: dict[int, FrameList] = field(default_factory=dict)  # type 1
    dump: dict[int, FrameList] = field(default_factory=dict)  # type 12? / sub cmd 4
    svg: dict[int, FrameList] = field(default_factory=dict)  # type 13
    line: dict[int, FrameList] = field(default_factory=dict)  # type 10 possibly breakpoint? / sub cmd 3
    visual_safety_zone: dict[int, FrameList] = field(default_factory=dict)  # type 25
    plan: dict[str, Plan] = field(default_factory=dict)
    area_name: list[AreaHashNameList] = field(default_factory=list)
    current_mow_path: dict[int, dict[int, MowPath]] = field(default_factory=dict)
    generated_geojson: dict[str, Any] = field(default_factory=dict)
    generated_mow_path_geojson: dict[str, Any] = field(default_factory=dict)

    def update_hash_lists(self, hashlist: list[int], bol_hash: int | None = None) -> None:
        if bol_hash:
            self.invalidate_maps(bol_hash)
        self.area = {hash_id: frames for hash_id, frames in self.area.items() if hash_id in hashlist}
        self.path = {hash_id: frames for hash_id, frames in self.path.items() if hash_id in hashlist}
        self.obstacle = {hash_id: frames for hash_id, frames in self.obstacle.items() if hash_id in hashlist}
        self.dump = {hash_id: frames for hash_id, frames in self.dump.items() if hash_id in hashlist}
        self.svg = {hash_id: frames for hash_id, frames in self.svg.items() if hash_id in hashlist}
        self.visual_safety_zone = {
            hash_id: frames for hash_id, frames in self.visual_safety_zone.items() if hash_id in hashlist
        }

        area_hashes = list(self.area.keys())
        for hash_id, plan_task in self.plan.copy().items():
            for item in plan_task.zone_hashs:
                if item not in area_hashes:
                    self.plan.pop(hash_id)
                    break

        self.area_name = [
            area_item
            for area_item in self.area_name
            if area_item.hash in self.area.keys() or area_item.hash in hashlist
        ]

    @property
    def hashlist(self) -> list[int]:
        if not self.root_hash_lists:
            return []
        # Combine data_couple from all RootHashLists
        return [i for root_list in self.root_hash_lists for obj in root_list.data for i in obj.data_couple]

    @property
    def area_root_hashlist(self) -> list[int]:
        if not self.root_hash_lists:
            return []
        # Combine data_couple from all RootHashLists
        return [
            i
            for root_list in self.root_hash_lists
            for obj in root_list.data
            for i in obj.data_couple
            if root_list.sub_cmd == 0
        ]

    def missing_hashlist(self, sub_cmd: int = 0) -> list[int]:
        """Return missing hashlist."""
        all_hash_ids = set(self.area.keys()).union(
            self.path.keys(), self.obstacle.keys(), self.dump.keys(), self.svg.keys(), self.visual_safety_zone.keys()
        )
        if sub_cmd == 3:
            all_hash_ids = set(self.line.keys())
        return [
            i
            for root_list in self.root_hash_lists
            for obj in root_list.data
            if root_list.sub_cmd == sub_cmd
            for i in obj.data_couple
            if i not in all_hash_ids
        ]

    def missing_root_hash_frame(self, hash_list: NavGetHashListAck) -> list[int]:
        """Return missing root hash frame."""
        target_root_list = next(
            (
                rhl
                for rhl in self.root_hash_lists
                if rhl.total_frame == hash_list.total_frame and rhl.sub_cmd == hash_list.sub_cmd
            ),
            None,
        )
        if target_root_list is None:
            return []

        return self.find_missing_frames(target_root_list)

    def update_root_hash_list(self, hash_list: NavGetHashListData) -> None:
        target_root_list = next(
            (
                rhl
                for rhl in self.root_hash_lists
                if rhl.total_frame == hash_list.total_frame and rhl.sub_cmd == hash_list.sub_cmd
            ),
            None,
        )

        if target_root_list is None:
            # Create new RootHashList if none exists for this total_frame
            new_root_list = RootHashList(total_frame=hash_list.total_frame, sub_cmd=hash_list.sub_cmd, data=[hash_list])
            self.root_hash_lists.append(new_root_list)
            return

        for index, obj in enumerate(target_root_list.data):
            if obj.current_frame == hash_list.current_frame:
                # Replace the item if current_frame matches
                target_root_list.data[index] = hash_list
                return

        # If no match was found, append the new item
        target_root_list.data.append(hash_list)

    def missing_hash_frame(self, hash_ack: NavGetHashListAck) -> list[int]:
        """Returns a combined list of all missing frames across all RootHashLists."""
        missing_frames = []
        filtered_lists = [rl for rl in self.root_hash_lists if rl.sub_cmd == hash_ack.sub_cmd]
        for root_list in filtered_lists:
            missing = self.find_missing_frames(root_list)
            if missing:
                missing_frames.extend(missing)
        return missing_frames

    def missing_frame(self, hash_data: NavGetCommDataAck | SvgMessageAckT) -> list[int]:
        frame_list = self._get_frame_list_by_type_and_hash(hash_data)
        return self.find_missing_frames(frame_list)

    def _get_frame_list_by_type_and_hash(self, hash_data: NavGetCommDataAck | SvgMessageAckT) -> FrameList | None:
        """Get the appropriate FrameList based on hash_data type and hash."""
        path_type_mapping = self._get_path_type_mapping()
        target_dict = path_type_mapping.get(hash_data.type)

        if target_dict is None:
            return None

        # Handle SvgMessage with data_hash attribute
        if isinstance(hash_data, SvgMessageAckT):
            return target_dict.get(hash_data.data_hash)

        # Handle NavGetCommDataAck with hash attribute
        return target_dict.get(hash_data.hash)

    def update_plan(self, plan: Plan) -> None:
        if plan.total_plan_num != 0:
            self.plan[plan.plan_id] = plan

    def _get_path_type_mapping(self) -> dict[int, dict[int, FrameList]]:
        """Return mapping of PathType to corresponding hash dictionary."""
        return {
            PathType.AREA: self.area,
            PathType.OBSTACLE: self.obstacle,
            PathType.PATH: self.path,
            PathType.LINE: self.line,
            PathType.DUMP: self.dump,
            PathType.SVG: self.svg,
            PathType.VISUAL_SAFETY_ZONE: self.visual_safety_zone,
        }

    def update(self, hash_data: NavGetCommData | SvgMessage) -> bool:
        """Update the map data."""

        if hash_data.type == PathType.AREA and isinstance(hash_data, NavGetCommData):
            existing_name = next((area for area in self.area_name if area.hash == hash_data.hash), None)
            if not existing_name:
                name = f"area {len(self.area_name)+1}"
                self.area_name.append(AreaHashNameList(name=name, hash=hash_data.hash))
            result = self._add_hash_data(self.area, hash_data)
            self.update_hash_lists(self.hashlist)
            return result

        path_type_mapping = self._get_path_type_mapping()
        target_dict = path_type_mapping.get(hash_data.type)

        if target_dict is not None:
            return self._add_hash_data(target_dict, hash_data)

        return False

    def find_missing_mow_path_frames(self) -> dict[int, list[int]]:
        """Find missing frames in current_mow_path grouped by transaction_id.

        Returns a mapping of transaction_id -> list of missing frame numbers.
        Only transaction_ids with at least one missing frame are included.
        """
        missing_frames: dict[int, list[int]] = {}

        if not self.current_mow_path:
            return missing_frames

        for transaction_id, frames_by_index in self.current_mow_path.items():
            if not frames_by_index:
                continue

            # Get total_frame from any MowPath object for this transaction_id
            any_mow_path = next(iter(frames_by_index.values()))
            total_frame = any_mow_path.total_frame

            if total_frame == 0:
                continue

            expected_frames = set(range(1, total_frame + 1))
            current_frames = set(frames_by_index.keys())
            missing_for_transaction = sorted(expected_frames - current_frames)

            if missing_for_transaction:
                missing_frames[transaction_id] = missing_for_transaction

        return missing_frames

    def update_mow_path(self, path: MowPath) -> None:
        """Update the current_mow_path with the latest MowPath data."""
        # TODO check if we need to clear the current_mow_path first
        transaction_id = path.transaction_id
        if transaction_id not in self.current_mow_path:
            self.current_mow_path[transaction_id] = {}
        self.current_mow_path[transaction_id][path.current_frame] = path

    @staticmethod
    def find_missing_frames(frame_list: FrameList | RootHashList | None) -> list[int]:
        if frame_list is None:
            return []

        if frame_list.total_frame == len(frame_list.data):
            return []
        number_list = list(range(1, frame_list.total_frame + 1))

        current_frames = {frame.current_frame for frame in frame_list.data}
        missing_numbers = [num for num in number_list if num not in current_frames]
        return missing_numbers

    @staticmethod
    def _add_hash_data(hash_dict: dict[int, FrameList], hash_data: NavGetCommData | SvgMessage) -> bool:
        if isinstance(hash_data, SvgMessage):
            if hash_dict.get(hash_data.data_hash, None) is None:
                hash_dict[hash_data.data_hash] = FrameList(total_frame=hash_data.total_frame, data=[hash_data])
                return True

            if hash_data not in hash_dict[hash_data.data_hash].data:
                exists = next(
                    (
                        rhl
                        for rhl in hash_dict[hash_data.data_hash].data
                        if rhl.current_frame == hash_data.current_frame
                    ),
                    None,
                )
                if exists:
                    return True
                hash_dict[hash_data.data_hash].data.append(hash_data)
                return True
            return False

        if hash_dict.get(hash_data.hash, None) is None:
            hash_dict[hash_data.hash] = FrameList(total_frame=hash_data.total_frame, data=[hash_data])
            return True

        if hash_data not in hash_dict[hash_data.hash].data:
            exists = next(
                (rhl for rhl in hash_dict[hash_data.hash].data if rhl.current_frame == hash_data.current_frame),
                None,
            )
            if exists:
                return True
            hash_dict[hash_data.hash].data.append(hash_data)
            return True
        return False

    def invalidate_maps(self, bol_hash: int) -> None:
        if MurMurHashUtil.hash_unsigned_list(self.area_root_hashlist) != bol_hash:
            self.root_hash_lists = []



from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Any

import betterproto2

from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model.device import MowingDevice
from pymammotion.data.model.raw_data import RawMowerData
from pymammotion.data.mower_state_manager import MowerStateManager
from pymammotion.proto import LubaMsg

_LOGGER = logging.getLogger(__name__)


class MammotionBaseDevice(ABC):
    """Base class for Mammotion devices."""

    def __init__(self, state_manager: MowerStateManager, cloud_device: Device) -> None:
        """Initialize MammotionBaseDevice."""
        self.loop = asyncio.get_event_loop()
        self._state_manager = state_manager
        self._raw_data = dict()
        self._raw_mower_data: RawMowerData = RawMowerData()
        self._notify_future: asyncio.Future[bytes] | None = None
        self._cloud_device = cloud_device

    def _update_raw_data(self, data: bytes) -> None:
        """Update raw and model data from notifications."""
        tmp_msg = LubaMsg().parse(data)
        res = betterproto2.which_one_of(tmp_msg, "LubaSubMsg")
        match res[0]:
            case "nav":
                self._update_nav_data(tmp_msg)
            case "sys":
                self._update_sys_data(tmp_msg)
            case "driver":
                self._update_driver_data(tmp_msg)
            case "net":
                self._update_net_data(tmp_msg)
            case "mul":
                self._update_mul_data(tmp_msg)
            case "ota":
                self._update_ota_data(tmp_msg)

        self._raw_mower_data.update_raw(self._raw_data)

    def _update_nav_data(self, tmp_msg) -> None:
        """Update navigation data."""
        nav_sub_msg = betterproto2.which_one_of(tmp_msg.nav, "SubNavMsg")
        if nav_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", nav_sub_msg[0])
            return
        nav = self._raw_data.get("nav", {})
        if isinstance(nav_sub_msg[1], int):
            nav[nav_sub_msg[0]] = nav_sub_msg[1]
        else:
            nav[nav_sub_msg[0]] = nav_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["nav"] = nav

    def _update_sys_data(self, tmp_msg: LubaMsg) -> None:
        """Update system data."""
        sys_sub_msg = betterproto2.which_one_of(tmp_msg.sys, "SubSysMsg")
        if sys_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", sys_sub_msg[0])
            return
        sys = self._raw_data.get("sys", {})
        sys[sys_sub_msg[0]] = sys_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["sys"] = sys

    def _update_driver_data(self, tmp_msg: LubaMsg) -> None:
        """Update driver data."""
        drv_sub_msg = betterproto2.which_one_of(tmp_msg.driver, "SubDrvMsg")
        if drv_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", drv_sub_msg[0])
            return
        drv = self._raw_data.get("driver", {})
        drv[drv_sub_msg[0]] = drv_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["driver"] = drv

    def _update_net_data(self, tmp_msg: LubaMsg) -> None:
        """Update network data."""
        net_sub_msg = betterproto2.which_one_of(tmp_msg.net, "NetSubType")
        if net_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", net_sub_msg[0])
            return
        net = self._raw_data.get("net", {})
        if isinstance(net_sub_msg[1], int):
            net[net_sub_msg[0]] = net_sub_msg[1]
        else:
            net[net_sub_msg[0]] = net_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["net"] = net

    def _update_mul_data(self, tmp_msg: LubaMsg) -> None:
        """Update mul data."""
        mul_sub_msg = betterproto2.which_one_of(tmp_msg.mul, "SubMul")
        if mul_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", mul_sub_msg[0])
            return
        mul = self._raw_data.get("mul", {})
        mul[mul_sub_msg[0]] = mul_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["mul"] = mul

    def _update_ota_data(self, tmp_msg: LubaMsg) -> None:
        """Update OTA data."""
        ota_sub_msg = betterproto2.which_one_of(tmp_msg.ota, "SubOtaMsg")
        if ota_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", ota_sub_msg[0])
            return
        ota = self._raw_data.get("ota", {})
        ota[ota_sub_msg[0]] = ota_sub_msg[1].to_dict(casing=betterproto2.Casing.SNAKE)
        self._raw_data["ota"] = ota

    @property
    def raw_data(self) -> dict[str, Any]:
        """Get the raw data of the device."""
        return self._raw_data

    @property
    def mower(self) -> MowingDevice:
        """Get the LubaMsg of the device."""
        return self._state_manager.get_device()

    @abstractmethod
    async def queue_command(self, key: str, **kwargs: Any) -> None:
        """Queue commands to mower."""

    @abstractmethod
    async def _ble_sync(self) -> None:
        """Send ble sync command every 3 seconds or sooner."""

    @abstractmethod
    def stop(self) -> None:
        """Stop everything ready for destroying."""

    async def async_read_settings(self) -> None:
        """Read settings from device."""
        # no cutting in rain nav_sys_param_cmd (id 3 context 1/0)
        await self.queue_command("read_write_device", rw_id=3, context=1, rw=0)
        # ??
        await self.queue_command("read_write_device", rw_id=4, context=1, rw=0)
        # turning mode nav_sys_param_cmd (id 6, context 1/0)
        await self.queue_command("read_write_device", rw_id=6, context=1, rw=0)
        # traversal mode
        await self.queue_command("read_write_device", rw_id=7, context=1, rw=0)

        await self.queue_command("read_and_set_sidelight", is_sidelight=True, operate=1)

        await self.queue_command("read_and_set_rtk_pairing_code", op=1, cfg="")

    async def async_get_errors(self) -> None:
        """Error codes."""
        await self.queue_command("read_write_device", rw_id=5, rw=1, context=2)
        await self.queue_command("read_write_device", rw_id=5, rw=1, context=3)

    async def command(self, key: str, **kwargs: Any) -> None:
        """Send a command to the device."""
        await self.queue_command(key, **kwargs)

    @property
    def state_manager(self) -> MowerStateManager:
        return self._state_manager



from dataclasses import dataclass
from enum import Enum
from typing import Annotated, Literal

from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Alias


@dataclass
class GroupIdListItem(DataClassORJSONMixin):
    groupId: str
    groupType: Literal["ISOLATION", "ILOP_APP"]


# are there other values?
class StatusType(Enum):
    CONNECTED = 1
    DISCONNECTED = 3


@dataclass
class Status(DataClassORJSONMixin):
    time: int
    value: StatusType


@dataclass
class Params(DataClassORJSONMixin):
    group_id_list: Annotated[list[GroupIdListItem], Alias("groupIdList")]
    net_type: Annotated[Literal["NET_WIFI", "NET_MNET"], Alias("netType")]
    active_time: Annotated[int, Alias("activeTime")]
    ip: str
    aliyun_commodity_code: Annotated[Literal["iothub_senior"], Alias("aliyunCommodityCode")]
    category_key: Annotated[Literal["LawnMower", "Tracker"], Alias("categoryKey")]
    node_type: Annotated[Literal["DEVICE"], Alias("nodeType")]
    product_key: Annotated[str, Alias("productKey")]
    status_last: Annotated[int, Alias("statusLast")]
    device_name: Annotated[str, Alias("deviceName")]
    iot_id: Annotated[str, Alias("iotId")]
    namespace: str
    tenant_id: Annotated[str, Alias("tenantId")]
    thing_type: Annotated[Literal["DEVICE"], Alias("thingType")]
    tenant_instance_id: Annotated[str, Alias("tenantInstanceId")]
    category_id: Annotated[int, Alias("categoryId")]
    status: Annotated[Status, Alias("status")]

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True


@dataclass
class ThingStatusMessage(DataClassORJSONMixin):
    method: Literal["thing.status"]
    id: str
    params: Params
    version: Literal["1.0"]



"""bidire_reqconver_path as a model."""

from dataclasses import dataclass, field

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class CurrentTaskSettings(DataClassORJSONMixin):
    pver: int = 0
    job_id: int = 0
    job_ver: int = 0
    job_mode: int = 0
    sub_cmd: int = 0
    edge_mode: int = 0
    knife_height: int = 0
    channel_width: int = 0
    ultra_wave: int = 0
    channel_mode: int = 0
    toward: int = 0
    speed: float = 0.0
    zone_hashs: list[int] = field(default_factory=list)
    path_hash: int = 0
    reserved: str = ""
    result: int = 0
    toward_mode: int = 0
    toward_included_angle: int = 0



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
IT_BASE: infoType
IT_OTA: infoType

class FotaInfo_t(_message.Message):
    __slots__ = ["need_ota_img_size", "need_ota_num", "ota_oid", "ota_otype", "ota_version"]
    NEED_OTA_IMG_SIZE_FIELD_NUMBER: _ClassVar[int]
    NEED_OTA_NUM_FIELD_NUMBER: _ClassVar[int]
    OTA_OID_FIELD_NUMBER: _ClassVar[int]
    OTA_OTYPE_FIELD_NUMBER: _ClassVar[int]
    OTA_VERSION_FIELD_NUMBER: _ClassVar[int]
    need_ota_img_size: int
    need_ota_num: int
    ota_oid: str
    ota_otype: int
    ota_version: str
    def __init__(self, need_ota_num: _Optional[int] = ..., need_ota_img_size: _Optional[int] = ..., ota_otype: _Optional[int] = ..., ota_oid: _Optional[str] = ..., ota_version: _Optional[str] = ...) -> None: ...

class FotaSubInfo_t(_message.Message):
    __slots__ = ["sub_img_size", "sub_img_url", "sub_mod_id", "sub_mod_ota_flag", "sub_mod_version"]
    SUB_IMG_SIZE_FIELD_NUMBER: _ClassVar[int]
    SUB_IMG_URL_FIELD_NUMBER: _ClassVar[int]
    SUB_MOD_ID_FIELD_NUMBER: _ClassVar[int]
    SUB_MOD_OTA_FLAG_FIELD_NUMBER: _ClassVar[int]
    SUB_MOD_VERSION_FIELD_NUMBER: _ClassVar[int]
    sub_img_size: int
    sub_img_url: str
    sub_mod_id: int
    sub_mod_ota_flag: int
    sub_mod_version: str
    def __init__(self, sub_mod_ota_flag: _Optional[int] = ..., sub_mod_id: _Optional[int] = ..., sub_img_size: _Optional[int] = ..., sub_mod_version: _Optional[str] = ..., sub_img_url: _Optional[str] = ...) -> None: ...

class MctlOta(_message.Message):
    __slots__ = ["fota_info", "fota_sub_info", "fw_download_ctrl", "toapp_get_info_rsp", "todev_get_info_req"]
    FOTA_INFO_FIELD_NUMBER: _ClassVar[int]
    FOTA_SUB_INFO_FIELD_NUMBER: _ClassVar[int]
    FW_DOWNLOAD_CTRL_FIELD_NUMBER: _ClassVar[int]
    TOAPP_GET_INFO_RSP_FIELD_NUMBER: _ClassVar[int]
    TODEV_GET_INFO_REQ_FIELD_NUMBER: _ClassVar[int]
    fota_info: FotaInfo_t
    fota_sub_info: FotaSubInfo_t
    fw_download_ctrl: fwDownloadCtrl
    toapp_get_info_rsp: getInfoRsp
    todev_get_info_req: getInfoReq
    def __init__(self, todev_get_info_req: _Optional[_Union[getInfoReq, _Mapping]] = ..., toapp_get_info_rsp: _Optional[_Union[getInfoRsp, _Mapping]] = ..., fw_download_ctrl: _Optional[_Union[fwDownloadCtrl, _Mapping]] = ..., fota_info: _Optional[_Union[FotaInfo_t, _Mapping]] = ..., fota_sub_info: _Optional[_Union[FotaSubInfo_t, _Mapping]] = ...) -> None: ...

class baseInfo(_message.Message):
    __slots__ = ["batt_val", "dev_status", "dev_version", "init_status", "is_tilt"]
    BATT_VAL_FIELD_NUMBER: _ClassVar[int]
    DEV_STATUS_FIELD_NUMBER: _ClassVar[int]
    DEV_VERSION_FIELD_NUMBER: _ClassVar[int]
    INIT_STATUS_FIELD_NUMBER: _ClassVar[int]
    IS_TILT_FIELD_NUMBER: _ClassVar[int]
    batt_val: int
    dev_status: int
    dev_version: str
    init_status: int
    is_tilt: int
    def __init__(self, dev_version: _Optional[str] = ..., dev_status: _Optional[int] = ..., batt_val: _Optional[int] = ..., init_status: _Optional[int] = ..., is_tilt: _Optional[int] = ...) -> None: ...

class downlink_t(_message.Message):
    __slots__ = ["data", "fw_id", "pkg_seq"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    FW_ID_FIELD_NUMBER: _ClassVar[int]
    PKG_SEQ_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[int]
    fw_id: int
    pkg_seq: int
    def __init__(self, fw_id: _Optional[int] = ..., pkg_seq: _Optional[int] = ..., data: _Optional[_Iterable[int]] = ...) -> None: ...

class fwDownloadCtrl(_message.Message):
    __slots__ = ["cmd", "downlink", "uplink"]
    CMD_FIELD_NUMBER: _ClassVar[int]
    DOWNLINK_FIELD_NUMBER: _ClassVar[int]
    UPLINK_FIELD_NUMBER: _ClassVar[int]
    cmd: int
    downlink: downlink_t
    uplink: uplink_t
    def __init__(self, uplink: _Optional[_Union[uplink_t, _Mapping]] = ..., downlink: _Optional[_Union[downlink_t, _Mapping]] = ..., cmd: _Optional[int] = ...) -> None: ...

class getInfoReq(_message.Message):
    __slots__ = ["type"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: infoType
    def __init__(self, type: _Optional[_Union[infoType, str]] = ...) -> None: ...

class getInfoRsp(_message.Message):
    __slots__ = ["base", "ota", "result", "type"]
    BASE_FIELD_NUMBER: _ClassVar[int]
    OTA_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    base: baseInfo
    ota: otaInfo
    result: int
    type: infoType
    def __init__(self, base: _Optional[_Union[baseInfo, _Mapping]] = ..., ota: _Optional[_Union[otaInfo, _Mapping]] = ..., result: _Optional[int] = ..., type: _Optional[_Union[infoType, str]] = ...) -> None: ...

class otaInfo(_message.Message):
    __slots__ = ["message", "otaid", "progress", "result", "version"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OTAID_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    message: str
    otaid: str
    progress: int
    result: int
    version: str
    def __init__(self, otaid: _Optional[str] = ..., version: _Optional[str] = ..., progress: _Optional[int] = ..., result: _Optional[int] = ..., message: _Optional[str] = ...) -> None: ...

class uplink_t(_message.Message):
    __slots__ = ["pkg_seq", "status"]
    PKG_SEQ_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    pkg_seq: int
    status: int
    def __init__(self, pkg_seq: _Optional[int] = ..., status: _Optional[int] = ...) -> None: ...

class infoType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from pymammotion.mammotion.commands.messages.driver import MessageDriver
from pymammotion.mammotion.commands.messages.media import MessageMedia
from pymammotion.mammotion.commands.messages.navigation import MessageNavigation
from pymammotion.mammotion.commands.messages.network import MessageNetwork
from pymammotion.mammotion.commands.messages.ota import MessageOta
from pymammotion.mammotion.commands.messages.system import MessageSystem
from pymammotion.mammotion.commands.messages.video import MessageVideo
from pymammotion.utility.device_type import DeviceType
from pymammotion.utility.movement import get_percent, transform_both_speeds


class MammotionCommand(
    MessageSystem, MessageNavigation, MessageNetwork, MessageOta, MessageVideo, MessageMedia, MessageDriver
):
    """MQTT commands for Luba."""

    def __init__(self, device_name: str, user_account: int) -> None:
        self._device_name = device_name
        self._product_key = ""
        self.user_account = user_account

    def get_device_name(self) -> str:
        """Get device name."""
        return self._device_name

    def read_write_device(self, rw_id: int, context: int, rw: int):
        if (
            rw_id == 6 or rw_id == 3 or rw_id == 7 or rw_id == 8 or rw_id == 10 or rw_id == 11
        ) and DeviceType.is_luba_pro(self.get_device_name()):
            return self.allpowerfull_rw_adapter_x3(rw_id, context, rw)
        return self.allpowerfull_rw(rw_id, context, rw)

    def traverse_mode(self, context: int) -> bytes:
        """Sets the traversal mode back to charger."""
        # setReChargeMode
        # 0 direct
        # 1 follow the perimeter
        return self.read_write_device(7, context, 1)

    def turning_mode(self, context: int) -> bytes:
        """Sets the traversal mode back to charger."""
        # setTurnAroundMode
        # 0 zero turn
        # 1 multipoint turn
        return self.read_write_device(6, context, 1)

    def get_error_code(self) -> bytes:
        return self.read_write_device(5, 2, 1)

    def get_error_timestamp(self) -> bytes:
        return self.read_write_device(5, 3, 1)

    def get_device_product_key(self) -> str:
        return self._product_key

    def set_device_product_key(self, product_key: str) -> None:
        self._product_key = product_key

    def move_forward(self, linear: float) -> bytes:
        """Move forward. values 0.0 1.0."""
        linear_percent = get_percent(abs(linear * 100))
        (linear_speed, angular_speed) = transform_both_speeds(90.0, 0.0, linear_percent, 0.0)
        return self.send_movement(linear_speed=linear_speed, angular_speed=angular_speed)

    def move_back(self, linear: float) -> bytes:
        """Move back. values 0.0 1.0."""
        linear_percent = get_percent(abs(linear * 100))
        (linear_speed, angular_speed) = transform_both_speeds(270.0, 0.0, linear_percent, 0.0)
        return self.send_movement(linear_speed=linear_speed, angular_speed=angular_speed)

    def move_left(self, angular: float) -> bytes:
        """Move forward. values 0.0 1.0."""
        angular_percent = get_percent(abs(angular * 100))
        (linear_speed, angular_speed) = transform_both_speeds(0.0, 180.0, 0.0, angular_percent)
        return self.send_movement(linear_speed=linear_speed, angular_speed=angular_speed)

    def move_right(self, angular: float) -> bytes:
        """Move back. values 0.0 1.0."""
        angular_percent = get_percent(abs(angular * 100))
        (linear_speed, angular_speed) = transform_both_speeds(0.0, 0.0, 0.0, angular_percent)
        return self.send_movement(linear_speed=linear_speed, angular_speed=angular_speed)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

ACTIVATION_FAILED: MUL_VIDEO_ERROR_CODE
ALL: MUL_CAMERA_POSITION
CREATE_CHANNEL_FAILED: MUL_VIDEO_ERROR_CODE
DESCRIPTOR: _descriptor.FileDescriptor
DUTCH: MUL_LANGUAGE
ENGLISH: MUL_LANGUAGE
FRENCH: MUL_LANGUAGE
GERMAN: MUL_LANGUAGE
HW_ERROR: MUL_WIPER_ERROR_CODE
ITALIAN: MUL_LANGUAGE
LEFT: MUL_CAMERA_POSITION
MAN: MUL_SEX
NAVIGATION_WORK_FORBID: MUL_WIPER_ERROR_CODE
NETWORK_NOT_AVAILABLE: MUL_VIDEO_ERROR_CODE
NONE_LAN: MUL_LANGUAGE
NONE_SEX: MUL_SEX
PARAM_INVAILD: MUL_VIDEO_ERROR_CODE
PORTUGUESE: MUL_LANGUAGE
REAR: MUL_CAMERA_POSITION
RIGHT: MUL_CAMERA_POSITION
SET_SUCCESS: MUL_WIPER_ERROR_CODE
SPANISH: MUL_LANGUAGE
SUCCESS: MUL_VIDEO_ERROR_CODE
SWEDISH: MUL_LANGUAGE
WOMAN: MUL_SEX
manual_power_off: lamp_manual_ctrl_sta
manual_power_on: lamp_manual_ctrl_sta
power_ctrl_on: lamp_ctrl_sta
power_off: lamp_ctrl_sta
power_on: lamp_ctrl_sta

class GetHeadlamp(_message.Message):
    __slots__ = ["get_ids"]
    GET_IDS_FIELD_NUMBER: _ClassVar[int]
    get_ids: int
    def __init__(self, get_ids: _Optional[int] = ...) -> None: ...

class Getlamprsp(_message.Message):
    __slots__ = ["get_ids", "lamp_bright", "lamp_ctrl", "lamp_manual_ctrl", "result"]
    GET_IDS_FIELD_NUMBER: _ClassVar[int]
    LAMP_BRIGHT_FIELD_NUMBER: _ClassVar[int]
    LAMP_CTRL_FIELD_NUMBER: _ClassVar[int]
    LAMP_MANUAL_CTRL_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    get_ids: int
    lamp_bright: int
    lamp_ctrl: lamp_ctrl_sta
    lamp_manual_ctrl: lamp_manual_ctrl_sta
    result: int
    def __init__(self, get_ids: _Optional[int] = ..., result: _Optional[int] = ..., lamp_ctrl: _Optional[_Union[lamp_ctrl_sta, str]] = ..., lamp_bright: _Optional[int] = ..., lamp_manual_ctrl: _Optional[_Union[lamp_manual_ctrl_sta, str]] = ...) -> None: ...

class MulAudioCfg(_message.Message):
    __slots__ = ["au_language", "au_switch", "sex"]
    AU_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    AU_SWITCH_FIELD_NUMBER: _ClassVar[int]
    SEX_FIELD_NUMBER: _ClassVar[int]
    au_language: MUL_LANGUAGE
    au_switch: int
    sex: MUL_SEX
    def __init__(self, au_switch: _Optional[int] = ..., au_language: _Optional[_Union[MUL_LANGUAGE, str]] = ..., sex: _Optional[_Union[MUL_SEX, str]] = ...) -> None: ...

class MulSetAudio(_message.Message):
    __slots__ = ["at_switch", "au_language", "sex"]
    AT_SWITCH_FIELD_NUMBER: _ClassVar[int]
    AU_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    SEX_FIELD_NUMBER: _ClassVar[int]
    at_switch: int
    au_language: MUL_LANGUAGE
    sex: MUL_SEX
    def __init__(self, at_switch: _Optional[int] = ..., au_language: _Optional[_Union[MUL_LANGUAGE, str]] = ..., sex: _Optional[_Union[MUL_SEX, str]] = ...) -> None: ...

class MulSetEncode(_message.Message):
    __slots__ = ["encode"]
    ENCODE_FIELD_NUMBER: _ClassVar[int]
    encode: bool
    def __init__(self, encode: bool = ...) -> None: ...

class MulSetVideo(_message.Message):
    __slots__ = ["position", "vi_switch"]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    VI_SWITCH_FIELD_NUMBER: _ClassVar[int]
    position: MUL_CAMERA_POSITION
    vi_switch: int
    def __init__(self, position: _Optional[_Union[MUL_CAMERA_POSITION, str]] = ..., vi_switch: _Optional[int] = ...) -> None: ...

class MulSetVideoAck(_message.Message):
    __slots__ = ["error_code"]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    error_code: MUL_VIDEO_ERROR_CODE
    def __init__(self, error_code: _Optional[_Union[MUL_VIDEO_ERROR_CODE, str]] = ...) -> None: ...

class MulSetWiper(_message.Message):
    __slots__ = ["round"]
    ROUND_FIELD_NUMBER: _ClassVar[int]
    round: int
    def __init__(self, round: _Optional[int] = ...) -> None: ...

class MulSetWiperAck(_message.Message):
    __slots__ = ["error_code"]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    error_code: MUL_WIPER_ERROR_CODE
    def __init__(self, error_code: _Optional[_Union[MUL_WIPER_ERROR_CODE, str]] = ...) -> None: ...

class SetHeadlamp(_message.Message):
    __slots__ = ["ctrl_lamp_bright", "lamp_bright", "lamp_ctrl", "lamp_manual_ctrl", "lamp_power_ctrl", "set_ids"]
    CTRL_LAMP_BRIGHT_FIELD_NUMBER: _ClassVar[int]
    LAMP_BRIGHT_FIELD_NUMBER: _ClassVar[int]
    LAMP_CTRL_FIELD_NUMBER: _ClassVar[int]
    LAMP_MANUAL_CTRL_FIELD_NUMBER: _ClassVar[int]
    LAMP_POWER_CTRL_FIELD_NUMBER: _ClassVar[int]
    SET_IDS_FIELD_NUMBER: _ClassVar[int]
    ctrl_lamp_bright: bool
    lamp_bright: int
    lamp_ctrl: lamp_ctrl_sta
    lamp_manual_ctrl: lamp_manual_ctrl_sta
    lamp_power_ctrl: int
    set_ids: int
    def __init__(self, set_ids: _Optional[int] = ..., lamp_power_ctrl: _Optional[int] = ..., lamp_ctrl: _Optional[_Union[lamp_ctrl_sta, str]] = ..., ctrl_lamp_bright: bool = ..., lamp_bright: _Optional[int] = ..., lamp_manual_ctrl: _Optional[_Union[lamp_manual_ctrl_sta, str]] = ...) -> None: ...

class Setlamprsp(_message.Message):
    __slots__ = ["result", "set_ids"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    SET_IDS_FIELD_NUMBER: _ClassVar[int]
    result: int
    set_ids: int
    def __init__(self, set_ids: _Optional[int] = ..., result: _Optional[int] = ...) -> None: ...

class SocMul(_message.Message):
    __slots__ = ["audio_cfg", "get_lamp", "get_lamp_rsp", "req_encode", "set_audio", "set_lamp", "set_lamp_rsp", "set_video", "set_video_ack", "set_wiper", "set_wiper_ack"]
    AUDIO_CFG_FIELD_NUMBER: _ClassVar[int]
    GET_LAMP_FIELD_NUMBER: _ClassVar[int]
    GET_LAMP_RSP_FIELD_NUMBER: _ClassVar[int]
    REQ_ENCODE_FIELD_NUMBER: _ClassVar[int]
    SET_AUDIO_FIELD_NUMBER: _ClassVar[int]
    SET_LAMP_FIELD_NUMBER: _ClassVar[int]
    SET_LAMP_RSP_FIELD_NUMBER: _ClassVar[int]
    SET_VIDEO_ACK_FIELD_NUMBER: _ClassVar[int]
    SET_VIDEO_FIELD_NUMBER: _ClassVar[int]
    SET_WIPER_ACK_FIELD_NUMBER: _ClassVar[int]
    SET_WIPER_FIELD_NUMBER: _ClassVar[int]
    audio_cfg: MulAudioCfg
    get_lamp: GetHeadlamp
    get_lamp_rsp: Getlamprsp
    req_encode: MulSetEncode
    set_audio: MulSetAudio
    set_lamp: SetHeadlamp
    set_lamp_rsp: Setlamprsp
    set_video: MulSetVideo
    set_video_ack: MulSetVideoAck
    set_wiper: MulSetWiper
    set_wiper_ack: MulSetWiperAck
    def __init__(self, set_audio: _Optional[_Union[MulSetAudio, _Mapping]] = ..., audio_cfg: _Optional[_Union[MulAudioCfg, _Mapping]] = ..., set_video: _Optional[_Union[MulSetVideo, _Mapping]] = ..., set_video_ack: _Optional[_Union[MulSetVideoAck, _Mapping]] = ..., set_wiper: _Optional[_Union[MulSetWiper, _Mapping]] = ..., set_wiper_ack: _Optional[_Union[MulSetWiperAck, _Mapping]] = ..., get_lamp: _Optional[_Union[GetHeadlamp, _Mapping]] = ..., set_lamp: _Optional[_Union[SetHeadlamp, _Mapping]] = ..., set_lamp_rsp: _Optional[_Union[Setlamprsp, _Mapping]] = ..., get_lamp_rsp: _Optional[_Union[Getlamprsp, _Mapping]] = ..., req_encode: _Optional[_Union[MulSetEncode, _Mapping]] = ...) -> None: ...

class MUL_LANGUAGE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_SEX(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_CAMERA_POSITION(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_VIDEO_ERROR_CODE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_WIPER_ERROR_CODE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class lamp_ctrl_sta(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class lamp_manual_ctrl_sta(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from __future__ import annotations

import base64
from collections.abc import Awaitable, Callable
import csv
from functools import wraps
import hashlib
import hmac
import json
import random
import time
from typing import Any, TypeVar, cast

from aiohttp import ClientSession
import jwt

from pymammotion.const import (
    MAMMOTION_API_DOMAIN,
    MAMMOTION_CLIENT_ID,
    MAMMOTION_CLIENT_SECRET,
    MAMMOTION_DOMAIN,
    MAMMOTION_OUATH2_CLIENT_ID,
    MAMMOTION_OUATH2_CLIENT_SECRET,
)
from pymammotion.http.encryption import EncryptionUtils
from pymammotion.http.model.camera_stream import StreamSubscriptionResponse, VideoResourceResponse
from pymammotion.http.model.http import (
    CheckDeviceVersion,
    DeviceInfo,
    DeviceRecords,
    ErrorInfo,
    JWTTokenInfo,
    LoginResponseData,
    MQTTConnection,
    Response,
    UnauthorizedException,
)
from pymammotion.http.model.response_factory import response_factory
from pymammotion.http.model.rtk import RTK

T = TypeVar("T")


def sign_with_hmac_sha256(data: str, app_secret: str) -> str:
    """Sign data with HMAC-SHA256 algorithm.

    Args:
        data: The data to sign
        app_secret: The secret key for signing

    Returns:
        Hex string of the signature

    Raises:
        RuntimeError: If signing fails

    """
    if data is None:
        raise ValueError("data cannot be None")
    if app_secret is None:
        raise ValueError("app_secret cannot be None")

    try:
        # Convert strings to bytes using UTF-8 encoding
        data_bytes = data.encode("utf-8")
        secret_bytes = app_secret.encode("utf-8")

        # Create HMAC-SHA256 hash
        hmac_obj = hmac.new(secret_bytes, data_bytes, hashlib.sha256)

        # Get the digest
        digest = hmac_obj.digest()

        # Convert to hex string
        hex_string = digest.hex()

        return hex_string

    except Exception as e:
        raise RuntimeError(f"toSignWithHmacSha256 error: {e}") from e


def create_oauth_signature(login_req: dict, client_id: str, client_secret: str, token_endpoint: str) -> str:
    """Create OAuth signature for login request.

    Args:
        login_req: Login request data as dictionary
        client_id: OAuth client ID
        client_secret: OAuth client secret
        token_endpoint: Token endpoint path

    Returns:
        HMAC-SHA256 signature

    """
    # Convert dict to JSON without HTML escaping (ensure_ascii=False handles this)
    json_data = json.dumps(login_req, ensure_ascii=False, separators=(",", ":"))

    # Get current timestamp in milliseconds
    timestamp = str(int(time.time() * 1000))

    # Construct the string to sign
    str_to_sign = f"{client_id}{timestamp}{token_endpoint}{json_data}"

    # Create MD5 hash of client secret
    try:
        md5_hash = hashlib.md5(client_secret.encode("utf-8")).digest()
        # Convert to hex string
        hashed_secret = md5_hash.hex()
    except Exception:
        hashed_secret = ""

    # Sign with HMAC-SHA256
    signature = sign_with_hmac_sha256(str_to_sign, hashed_secret)

    return signature


class MammotionHTTP:
    def __init__(
        self, account: str | None = None, password: str | None = None, session: ClientSession | None = None
    ) -> None:
        self.device_info: list[DeviceInfo] = []
        self.mqtt_credentials: MQTTConnection | None = None
        self.device_records: DeviceRecords = DeviceRecords(records=[], current=0, total=0, size=0, pages=0)
        self.expires_in = 0.0
        self.code = 0
        self.msg = None
        self._session = session if session else ClientSession()
        self.account = account
        self._password = password
        self._response: Response | None = None
        self.login_info: LoginResponseData | None = None
        self.jwt_info: JWTTokenInfo = JWTTokenInfo("", "")
        self._headers = {"User-Agent": "okhttp/4.9.3", "App-Version": "Home Assistant,1.15.6.14"}
        self.encryption_utils = EncryptionUtils()

        # Add this method to generate a 10-digit random number
        def get_10_random() -> str:
            """Generate a 10-digit random number as a string."""
            return "".join([str(random.randint(0, 9)) for _ in range(7)])

        # Replace the line in the __init__ method with:
        self.client_id = f"{int(time.time() * 1000)}_{get_10_random()}_1"

    @property
    def response(self) -> Response | None:
        return self._response

    @response.setter
    def response(self, response: Response) -> None:
        self._response = response
        decoded_token = jwt.decode(response.data.access_token, options={"verify_signature": False})
        if isinstance(decoded_token, dict):
            self.jwt_info = JWTTokenInfo(iot=decoded_token.get("iot", ""), robot=decoded_token.get("robot", ""))

    @staticmethod
    def generate_headers(token: str) -> dict:
        return {"Authorization": f"Bearer {token}"}

    @staticmethod
    def refresh_token_decorator(func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
        """Decorator to handle token refresh before executing a function.

        Args:
            func: The async function to be decorated

        Returns:
            The wrapped async function that handles token refresh

        """

        @wraps(func)
        async def wrapper(self: MammotionHTTP, *args: Any, **kwargs: Any) -> T:
            # Check if token will expire in the next 5 minutes
            if self.expires_in < time.time() + 300:  # 300 seconds = 5 minutes
                await self.refresh_login()
            return await func(self, *args, **kwargs)

        return wrapper

    async def handle_expiry(self, resp: Response) -> Response:
        if resp.code == 401 and self.account and self._password:
            return await self.login_v2(self.account, self._password)
        return resp

    async def login_by_email(self, email: str, password: str) -> Response[LoginResponseData]:
        return await self.login_v2(email, password)

    @refresh_token_decorator
    async def get_all_error_codes(self) -> dict[str, ErrorInfo]:
        """Retrieves and parses all error codes from the MAMMOTION API."""
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/user-server/v1/code/record/export-data",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        data = await resp.json()
        reader = csv.DictReader(data.get("data", "").split("\n"), delimiter=",")
        codes = dict()
        for row in reader:
            error_info = ErrorInfo(**cast(dict[str, Any], row))
            codes[error_info.code] = error_info
        return codes

    async def oauth_check(self) -> Response:
        """Check if token is valid.

        Returns 401 if token is invalid. We then need to re-authenticate, can try to refresh token first
        """
        resp = await self._session.post(
            f"{MAMMOTION_DOMAIN}/user-server/v1/user/oauth/check",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        data = await resp.json()
        return Response.from_dict(data)

    @refresh_token_decorator
    async def refresh_authorization_code(self) -> Response:
        """Refresh token."""
        resp = await self._session.post(
            f"{MAMMOTION_DOMAIN}/authorization/code",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
            json={"clientId": MAMMOTION_CLIENT_ID},
        )
        data = await resp.json()
        print(data)
        self.login_info.access_token = data["data"].get("accessToken", self.login_info.access_token)
        self.login_info.authorization_code = data["data"].get("code", self.login_info.authorization_code)
        await self.get_mqtt_credentials()
        return Response.from_dict(data)

    @refresh_token_decorator
    async def pair_devices_mqtt(self, mower_name: str, rtk_name: str) -> Response:
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/iot/device/pairing",
            headers=self._headers,
            json={"mowerName": mower_name, "rtkName": rtk_name},
        )
        data = await resp.json()
        if data.get("status") == 200:
            print(data)
            return Response.from_dict(data)
        else:
            print(data)
            return Response.from_dict(data)

    @refresh_token_decorator
    async def unpair_devices_mqtt(self, mower_name: str, rtk_name: str) -> Response:
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/iot/device/unpairing",
            headers=self._headers,
            json={"mowerName": mower_name, "rtkName": rtk_name},
        )
        data = await resp.json()
        if data.get("status") == 200:
            print(data)
            return Response.from_dict(data)
        else:
            print(data)
            return Response.from_dict(data)

    @refresh_token_decorator
    async def net_rtk_enable(self, device_id: str) -> Response:
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/iot/net-rtk/enable",
            headers=self._headers,
            json={"deviceId": device_id},
        )
        data = await resp.json()
        if data.get("status") == 200:
            print(data)
            return Response.from_dict(data)
        else:
            print(data)
            return Response.from_dict(data)

    @refresh_token_decorator
    async def get_stream_subscription(self, iot_id: str, is_yuka: bool) -> Response[StreamSubscriptionResponse]:
        # Prepare the payload with cameraStates based on is_yuka flag
        """Fetches stream subscription data for a given IoT device."""

        payload = {"deviceId": iot_id, "mode": 0, "cameraStates": []}

        # Add appropriate cameraStates based on the is_yuka flag
        # yukas have two cameras you could view [{"cameraState": 1}, {"cameraState": 0}, {"cameraState": 1}]
        # but its not useful so ignore this and only subscribe to the front one.
        if is_yuka:
            payload["cameraStates"] = [{"cameraState": 1}, {"cameraState": 0}, {"cameraState": 0}]
        else:
            payload["cameraStates"] = [{"cameraState": 1}, {"cameraState": 0}, {"cameraState": 0}]

        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/stream/token",
            json=payload,
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        data = await resp.json()
        # TODO catch errors from mismatch like token expire etc
        # Assuming the data format matches the expected structure
        response = Response[StreamSubscriptionResponse].from_dict(data)
        await self.handle_expiry(response)
        if response.code != 0:
            return response
        response.data = StreamSubscriptionResponse.from_dict(data.get("data", {}))
        return response

    @refresh_token_decorator
    async def get_video_resource(self, iot_id: str) -> Response[VideoResourceResponse]:
        """Fetch video resource for a given IoT ID."""
        resp = await self._session.get(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/video-resource/{iot_id}",
            headers={
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        data = await resp.json()
        # TODO catch errors from mismatch like token expire etc
        # Assuming the data format matches the expected structure
        response = Response[VideoResourceResponse].from_dict(data)
        if response.code != 0:
            return response
        response.data = VideoResourceResponse.from_dict(data.get("data", {}))
        return response

    @refresh_token_decorator
    async def get_device_ota_firmware(self, iot_ids: list[str]) -> Response[list[CheckDeviceVersion]]:
        """Checks device firmware versions for a list of IoT IDs."""
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/devices/version/check",
            json={"deviceIds": iot_ids},
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
        )
        data = await resp.json()
        # TODO catch errors from mismatch like token expire etc
        # Assuming the data format matches the expected structure
        return response_factory(Response[list[CheckDeviceVersion]], data)

    @refresh_token_decorator
    async def start_ota_upgrade(self, iot_id: str, version: str) -> Response[str]:
        """Initiates an OTA upgrade for a device."""
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/ota/device/upgrade",
            json={"deviceId": iot_id, "version": version},
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
        )
        data = await resp.json()
        # TODO catch errors from mismatch like token expire etc
        # Assuming the data format matches the expected structure
        return response_factory(Response[str], data)

    @refresh_token_decorator
    async def get_rtk_devices(self) -> Response[list[RTK]]:
        """Fetches stream subscription data from agora.io for a given IoT device."""
        resp = await self._session.get(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/rtk/devices",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        data = await resp.json()

        return response_factory(Response[list[RTK]], data)

    @refresh_token_decorator
    async def get_user_device_list(self) -> Response[list[DeviceInfo]]:
        """Fetches device list for a user (owned not shared, shared returns nothing)."""
        resp = await self._session.get(
            f"{MAMMOTION_API_DOMAIN}/device-server/v1/device/list",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
        )
        resp_dict = await resp.json()
        response = response_factory(Response[list[DeviceInfo]], resp_dict)
        self.device_info = response.data if response.data else self.device_info
        return response

    @refresh_token_decorator
    async def get_user_shared_device_page(self) -> Response[DeviceRecords]:
        """Fetches device list for a user (shared) but not accepted."""
        """Can set owned to zero or one to possibly check for not accepted mowers?"""
        resp = await self._session.post(
            f"{MAMMOTION_API_DOMAIN}/user-server/v1/share/device/page",
            json={"iotId": "", "owned": 0, "pageNumber": 1, "pageSize": 200, "statusList": [-1]},
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        resp_dict = await resp.json()
        response = response_factory(Response[DeviceRecords], resp_dict)
        self.devices_shared_info = response.data if response.data else self.devices_shared_info
        return response

    @refresh_token_decorator
    async def get_user_device_page(self) -> Response[DeviceRecords]:
        """Fetches device list for a user, is either new API or for newer devices."""
        resp = await self._session.post(
            f"{self.jwt_info.iot}/v1/user/device/page",
            json={
                "iotId": "",
                "pageNumber": 1,
                "pageSize": 100,
            },
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
        )
        if resp.status != 200:
            return Response.from_dict({"code": resp.status, "msg": "get device list failed"})
        resp_dict = await resp.json()
        response = response_factory(Response[DeviceRecords], resp_dict)
        self.device_records = response.data if response.data else self.device_records
        return response

    @refresh_token_decorator
    async def get_mqtt_credentials(self) -> Response[MQTTConnection]:
        """Get mammotion mqtt credentials"""
        resp = await self._session.post(
            f"{self.jwt_info.iot}/v1/mqtt/auth/jwt",
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
            },
        )
        if resp.status != 200:
            return Response.from_dict({"code": resp.status, "msg": "get mqtt failed"})
        resp_dict = await resp.json()
        response = response_factory(Response[MQTTConnection], resp_dict)
        self.mqtt_credentials = response.data
        return response

    @refresh_token_decorator
    async def mqtt_invoke(self, content: str, device_name: str, iot_id: str) -> Response[dict]:
        """Send mqtt commands to devices."""
        resp = await self._session.post(
            f"{self.jwt_info.iot}/v1/mqtt/rpc/thing/service/invoke",
            json={
                "args": {"content": content},
                "deviceName": device_name,
                "identifier": "device_protobuf_sync_service",
                "iotId": iot_id,
                "productKey": "",
            },
            headers={
                **self._headers,
                "Authorization": f"Bearer {self.login_info.access_token}",
                "Content-Type": "application/json",
                "User-Agent": "okhttp/4.9.3",
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
        )
        if resp.status != 200:
            return Response.from_dict({"code": resp.status, "msg": "invoke mqtt failed"})
        if resp.status == 401:
            raise UnauthorizedException("Access Token expired")
        resp_dict = await resp.json()
        return response_factory(Response[dict], resp_dict)

    async def refresh_login(self) -> Response[LoginResponseData]:
        if self.expires_in > time.time():
            res = await self.refresh_token_v2()
            if res.code == 0:
                return res
        return await self.login_v2(self.account, self._password)

    async def login(self, account: str, password: str) -> Response[LoginResponseData]:
        """Logs in to the service using provided account and password."""
        self.account = account
        self._password = password
        resp = await self._session.post(
            f"{MAMMOTION_DOMAIN}/oauth/token",
            headers={
                **self._headers,
                "Encrypt-Key": self.encryption_utils.encrypt_by_public_key(),
                "Decrypt-Type": "3",
                "Ec-Version": "v1",
            },
            params={
                "username": self.encryption_utils.encryption_by_aes(account),
                "password": self.encryption_utils.encryption_by_aes(password),
                "client_id": self.encryption_utils.encryption_by_aes(MAMMOTION_CLIENT_ID),
                "client_secret": self.encryption_utils.encryption_by_aes(MAMMOTION_CLIENT_SECRET),
                "grant_type": self.encryption_utils.encryption_by_aes("password"),
            },
        )
        if resp.status != 200:
            print(resp.json())
            return Response.from_dict({"code": resp.status, "msg": "Login failed"})
        data = await resp.json()
        login_response = response_factory(Response[LoginResponseData], data)
        if login_response is None or login_response.data is None:
            print(login_response)
            return Response.from_dict({"code": resp.status, "msg": "Login failed"})
        self.login_info = login_response.data
        self.expires_in = login_response.data.expires_in + time.time()
        self._headers["Authorization"] = f"Bearer {self.login_info.access_token}" if login_response.data else None
        self.response = login_response
        self.msg = login_response.msg
        self.code = login_response.code
        # TODO catch errors from mismatch user / password elsewhere
        # Assuming the data format matches the expected structure
        return login_response

    async def refresh_token_v2(self) -> Response[LoginResponseData]:
        """Refresh token v2."""

        refresh_request = {
            "client_id": MAMMOTION_OUATH2_CLIENT_ID,
            "refresh_token": self.login_info.refresh_token,
            "grant_type": "refresh_token",
        }

        oauth_signature = create_oauth_signature(
            login_req=refresh_request,
            client_id=MAMMOTION_OUATH2_CLIENT_ID,
            client_secret=MAMMOTION_OUATH2_CLIENT_SECRET,
            token_endpoint="/oauth2/token",
        )

        resp = await self._session.post(
            f"{MAMMOTION_DOMAIN}/oauth2/token",
            headers={
                **self._headers,
                "Ma-Iot-Signature": oauth_signature,
                "Ma-Timestamp": str(int(time.time())),
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
            params={
                **refresh_request,
            },
        )
        data = await resp.json()
        refresh_response = response_factory(Response[LoginResponseData], data)
        if refresh_response is None or refresh_response.data is None:
            return Response.from_dict({"code": resp.status, "msg": "Login failed"})
        self.login_info = refresh_response.data
        self.expires_in = refresh_response.data.expires_in + time.time()
        self._headers["Authorization"] = f"Bearer {self.login_info.access_token}" if refresh_response.data else None
        self.response = refresh_response
        self.msg = refresh_response.msg
        self.code = refresh_response.code
        return refresh_response

    async def login_v2(self, account: str, password: str) -> Response[LoginResponseData]:
        """Logs in to the service using provided account and password."""
        self.account = account
        self._password = password

        login_request = {
            "username": account,
            "password": base64.b64encode(password.encode("utf-8")).decode("utf-8"),
            "client_id": MAMMOTION_OUATH2_CLIENT_ID,
            "grant_type": "password",
            "authType": "0",
        }

        oauth_signature = create_oauth_signature(
            login_req=login_request,
            client_id=MAMMOTION_OUATH2_CLIENT_ID,
            client_secret=MAMMOTION_OUATH2_CLIENT_SECRET,
            token_endpoint="/oauth2/token",
        )

        resp = await self._session.post(
            f"{MAMMOTION_DOMAIN}/oauth2/token",
            headers={
                **self._headers,
                "Ma-App-Key": MAMMOTION_OUATH2_CLIENT_ID,
                "Ma-Signature": oauth_signature,
                "Ma-Timestamp": str(int(time.time())),
                "Client-Id": self.client_id,
                "Client-Type": "1",
            },
            params={
                **login_request,
            },
        )
        if resp.status != 200:
            return Response.from_dict({"code": resp.status, "msg": "Login failed"})
        data = await resp.json()
        if data.get("code") != 0:
            return Response.from_dict({"code": resp.status, "msg": data.get("msg") or "Login failed"})
        login_response = response_factory(Response[LoginResponseData], data)
        if login_response is None or login_response.data is None:
            return Response.from_dict({"code": resp.status, "msg": "Login failed"})
        self.login_info = login_response.data
        self.expires_in = login_response.data.expires_in + time.time()
        self._headers["Authorization"] = f"Bearer {self.login_info.access_token}" if login_response.data else None
        self.response = login_response
        self.msg = login_response.msg
        self.code = login_response.code
        # TODO catch errors from mismatch user / password elsewhere
        # Assuming the data format matches the expected structure
        return login_response



from dataclasses import dataclass
from typing import Annotated, Optional

from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Alias


@dataclass
class Device(DataClassORJSONMixin):
    """Unified device model supporting both Device and ShareNotification data"""

    # Core device fields (from Device model)
    gmt_modified: Annotated[int, Alias("gmtModified")]
    node_type: Annotated[str, Alias("nodeType")]
    device_name: Annotated[str, Alias("deviceName")]
    product_name: Annotated[str, Alias("productName")]
    status: int
    identity_id: Annotated[str, Alias("identityId")]

    # Required fields from original Device model
    net_type: Annotated[str, Alias("netType")]
    category_key: Annotated[str, Alias("categoryKey")]
    product_key: Annotated[str, Alias("productKey")]
    is_edge_gateway: Annotated[bool, Alias("isEdgeGateway")]
    category_name: Annotated[str, Alias("categoryName")]
    identity_alias: Annotated[str, Alias("identityAlias")]
    iot_id: Annotated[str, Alias("iotId")]
    bind_time: Annotated[int, Alias("bindTime")]
    owned: int
    thing_type: Annotated[str, Alias("thingType")]

    # Optional fields (common to both or nullable)
    nick_name: Annotated[Optional[str], Alias("nickName")] = None
    description: Optional[str] = None
    product_image: Annotated[Optional[str], Alias("productImage")] = None
    category_image: Annotated[Optional[str], Alias("categoryImage")] = None
    product_model: Annotated[Optional[str], Alias("productModel")] = None

    # Optional fields from ShareNotification only
    target_id: Annotated[Optional[str], Alias("targetId")] = None
    receiver_identity_id: Annotated[Optional[str], Alias("receiverIdentityId")] = None
    target_type: Annotated[Optional[str], Alias("targetType")] = None
    gmt_create: Annotated[Optional[int], Alias("gmtCreate")] = None
    batch_id: Annotated[Optional[str], Alias("batchId")] = None
    record_id: Annotated[Optional[str], Alias("recordId")] = None
    initiator_identity_id: Annotated[Optional[str], Alias("initiatorIdentityId")] = None
    is_receiver: Annotated[Optional[int], Alias("isReceiver")] = None
    initiator_alias: Annotated[Optional[str], Alias("initiatorAlias")] = None
    receiver_alias: Annotated[Optional[str], Alias("receiverAlias")] = None

    class Config(BaseConfig):
        omit_default = True
        allow_deserialization_not_by_alias = True


# # Alternative: Keep them separate but with a common base class
# @dataclass
# class BaseDevice(DataClassORJSONMixin):
#     """Base device model with common fields"""
#
#     gmt_modified: int
#     node_type: str
#     device_name: str
#     product_name: str
#     status: int
#     product_image: Optional[str] = None
#     category_image: Optional[str] = None
#     description: Optional[str] = None
#
#     class Config(BaseConfig):
#         omit_default = True
#         serialize_by_alias = True
#         aliases = {
#             "gmt_modified": "gmtModified",
#             "node_type": "nodeType",
#             "device_name": "deviceName",
#             "product_name": "productName",
#             "product_image": "productImage",
#             "category_image": "categoryImage",
#         }
#
#
# @dataclass
# class Device(BaseDevice):
#     """Full device model"""
#
#     net_type: str
#     category_key: str
#     product_key: str
#     is_edge_gateway: bool
#     category_name: str
#     identity_alias: str
#     iot_id: str
#     bind_time: int
#     owned: int
#     identity_id: str
#     thing_type: str
#     nick_name: Optional[str] = None
#     product_model: Optional[str] = None
#
#     class Config(BaseConfig):
#         omit_default = True
#         serialize_by_alias = True
#         aliases = {
#             **BaseDevice.Config.aliases,
#             "net_type": "netType",
#             "category_key": "categoryKey",
#             "product_key": "productKey",
#             "is_edge_gateway": "isEdgeGateway",
#             "category_name": "categoryName",
#             "identity_alias": "identityAlias",
#             "iot_id": "iotId",
#             "bind_time": "bindTime",
#             "identity_id": "identityId",
#             "thing_type": "thingType",
#             "nick_name": "nickName",
#             "product_model": "productModel",
#         }
#
#
# @dataclass
# class ShareNotification(BaseDevice):
#     """Share notification model extending base device"""
#
#     target_id: str
#     receiver_identity_id: str
#     target_type: str
#     gmt_create: int
#     batch_id: str
#     record_id: str
#     initiator_identity_id: str
#     is_receiver: int
#     initiator_alias: str
#     receiver_alias: str
#
#     # Optional fields that Device has but ShareNotification might not
#     net_type: Optional[str] = None
#     category_key: Optional[str] = None
#     product_key: Optional[str] = None
#     is_edge_gateway: Optional[bool] = None
#     category_name: Optional[str] = None
#     identity_alias: Optional[str] = None
#     iot_id: Optional[str] = None
#     bind_time: Optional[int] = None
#     owned: Optional[int] = None
#     identity_id: Optional[str] = None
#     thing_type: Optional[str] = None
#     nick_name: Optional[str] = None
#     product_model: Optional[str] = None
#
#     class Config(BaseConfig):
#         omit_default = True
#         serialize_by_alias = True
#         aliases = {
#             **BaseDevice.Config.aliases,
#             "target_id": "targetId",
#             "receiver_identity_id": "receiverIdentityId",
#             "target_type": "targetType",
#             "gmt_create": "gmtCreate",
#             "batch_id": "batchId",
#             "record_id": "recordId",
#             "initiator_identity_id": "initiatorIdentityId",
#             "is_receiver": "isReceiver",
#             "initiator_alias": "initiatorAlias",
#             "receiver_alias": "receiverAlias",
#             # Device fields that might be present
#             "net_type": "netType",
#             "category_key": "categoryKey",
#             "product_key": "productKey",
#             "is_edge_gateway": "isEdgeGateway",
#             "category_name": "categoryName",
#             "identity_alias": "identityAlias",
#             "iot_id": "iotId",
#             "bind_time": "bindTime",
#             "identity_id": "identityId",
#             "thing_type": "thingType",
#             "nick_name": "nickName",
#             "product_model": "productModel",
#         }


@dataclass
class Data(DataClassORJSONMixin):
    total: int
    data: list[Device]
    pageNo: int
    pageSize: int


@dataclass
class ListingDevAccountResponse(DataClassORJSONMixin):
    code: int
    data: Data | None
    id: str | None = None



import asyncio
from asyncio import InvalidStateError
import base64
from collections import deque
from collections.abc import Awaitable, Callable
import json
import logging
import time
from typing import Any

import betterproto2
from Tea.exceptions import UnretryableException

from pymammotion import AliyunMQTT, CloudIOTGateway, MammotionMQTT
from pymammotion.aliyun.cloud_gateway import DeviceOfflineException
from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.mower_state_manager import MowerStateManager
from pymammotion.data.mqtt.event import MammotionEventMessage, ThingEventMessage
from pymammotion.data.mqtt.properties import MammotionPropertiesMessage, ThingPropertiesMessage
from pymammotion.data.mqtt.status import ThingStatusMessage
from pymammotion.event.event import DataEvent
from pymammotion.mammotion.commands.mammotion_command import MammotionCommand
from pymammotion.mammotion.devices.base import MammotionBaseDevice
from pymammotion.proto import LubaMsg

_LOGGER = logging.getLogger(__name__)


class MammotionCloud:
    """Per account MQTT cloud."""

    def __init__(self, mqtt_client: AliyunMQTT | MammotionMQTT, cloud_client: CloudIOTGateway) -> None:
        """Initialize MammotionCloud."""
        self.cloud_client = cloud_client
        self.command_sent_time = 0
        self.loop = asyncio.get_event_loop()
        self.is_ready = False
        self.command_queue = asyncio.Queue()
        self._waiting_queue = deque()
        self.mqtt_message_event = DataEvent()
        self.mqtt_properties_event = DataEvent()
        self.mqtt_status_event = DataEvent()
        self.mqtt_device_event = DataEvent()
        self.on_ready_event = DataEvent()
        self.on_disconnected_event = DataEvent()
        self.on_connected_event = DataEvent()
        self._operation_lock = asyncio.Lock()
        self._mqtt_client = mqtt_client
        self._mqtt_client.on_connected = self.on_connected
        self._mqtt_client.on_disconnected = self.on_disconnected
        self._mqtt_client.on_message = self._on_mqtt_message
        self._mqtt_client.on_ready = self.on_ready

    async def on_ready(self) -> None:
        """Starts processing the queue and emits the ready event."""
        loop = asyncio.get_event_loop()
        loop.create_task(self.process_queue())
        await self.on_ready_event.data_event(None)

    def is_connected(self) -> bool:
        return self._mqtt_client.is_connected

    def disconnect(self) -> None:
        """Disconnect the MQTT client."""
        if self.is_connected:
            self._mqtt_client.disconnect()

    def connect_async(self) -> None:
        self._mqtt_client.connect_async()

    async def send_command(self, iot_id: str, command: bytes) -> None:
        await self._mqtt_client.send_cloud_command(iot_id, command)

    async def on_connected(self) -> None:
        """Callback for when MQTT connects."""
        await self.on_connected_event.data_event(None)

    async def on_disconnected(self) -> None:
        """Callback for when MQTT disconnects."""
        await self.on_disconnected_event.data_event(None)

    async def process_queue(self) -> None:
        while True:
            # Get the next item from the queue
            iot_id, key, command, future = await self.command_queue.get()
            try:
                # Process the command using _execute_command_locked
                result = await self._execute_command_locked(iot_id, key, command)
                # Set the result on the future
                future.set_result(result)
            except Exception as ex:
                # Set the exception on the future if something goes wrong
                try:
                    future.set_exception(ex)
                except InvalidStateError:
                    """Dead end, log an error."""
                    _LOGGER.exception("InvalidStateError while trying to bubble up exception")
            finally:
                # Mark the task as done
                self.command_queue.task_done()

    async def _execute_command_locked(self, iot_id: str, key: str, command: bytes) -> None:
        """Execute command and read response."""
        assert self._mqtt_client is not None
        self._key = key
        _LOGGER.debug("Sending command: %s", key)
        self.command_sent_time = time.time()
        await self._mqtt_client.send_cloud_command(iot_id, command)

    async def _on_mqtt_message(self, topic: str, payload: bytes, iot_id: str) -> None:
        """Handle incoming MQTT messages."""
        # _LOGGER.debug("MQTT message received on topic %s: %s, iot_id: %s", topic, payload, iot_id)
        json_str = payload.decode("utf-8")
        dict_payload = json.loads(json_str)
        await self._parse_mqtt_response(topic, dict_payload, iot_id)

    async def _parse_mqtt_response(self, topic: str, payload: dict, iot_id: str) -> None:
        """Parse and handle MQTT responses based on the topic.

        This function processes different types of MQTT messages received from various
        topics. It logs debug information and calls appropriate callback methods for
        each event type.

        Args:
            topic (str): The MQTT topic from which the message was received.
            payload (dict): The payload data of the MQTT message.

        """
        if topic.endswith("/app/down/account/bind_reply"):
            code = payload.get("code", 0)
            if code != 200:
                _LOGGER.error("Failed to bind account: %s", payload)
                # TODO send message to re login to aliyun mqtt
                self._disconnect_error = payload
                return

        if topic.endswith("/app/down/thing/events"):
            _LOGGER.debug("Thing event received")
            event = ThingEventMessage.from_dicts(payload)
            params = event.params
            if isinstance(params, dict) or params.identifier is None:
                _LOGGER.debug("Received dict params: %s", params)
                return
            if params.identifier == "device_protobuf_msg_event" and event.method == "thing.events":
                _LOGGER.debug("Protobuf event")
                # Call the callbacks for each cloudDevice
                await self.mqtt_message_event.data_event(event)
            if event.method == "thing.events":
                await self.mqtt_device_event.data_event(event)
            if event.method == "thing.properties":
                await self.mqtt_properties_event.data_event(event)
                _LOGGER.debug(event)
        elif topic.endswith("/app/down/thing/status"):
            status = ThingStatusMessage.from_dict(payload)
            await self.mqtt_status_event.data_event(status)
        elif topic.endswith("app/down/thing/properties"):
            property_event = ThingPropertiesMessage.from_dict(payload)
            await self.mqtt_properties_event.data_event(property_event)

        if topic.endswith("/thing/event/device_protobuf_msg_event/post"):
            _LOGGER.debug("Mammotion Thing event received")
            mammotion_event = MammotionEventMessage.from_dict(payload)
            mammotion_event.params.iot_id = iot_id
            await self.mqtt_message_event.data_event(mammotion_event)
        elif topic.endswith("/thing/event/property/post"):
            _LOGGER.debug("Mammotion Property event received")
            mammotion_property_event = MammotionPropertiesMessage.from_dict(payload)
            mammotion_property_event.params.iot_id = iot_id
            await self.mqtt_properties_event.data_event(mammotion_property_event)

    def _disconnect(self) -> None:
        """Disconnect the MQTT client."""
        self._mqtt_client.disconnect()

    @property
    def waiting_queue(self) -> deque:
        return self._waiting_queue


class MammotionBaseCloudDevice(MammotionBaseDevice):
    """Base class for Mammotion Cloud devices."""

    def __init__(self, mqtt: MammotionCloud, cloud_device: Device, state_manager: MowerStateManager) -> None:
        """Initialize MammotionBaseCloudDevice."""
        super().__init__(state_manager, cloud_device)
        self.stopped = False
        self.on_ready_callback: Callable[[], Awaitable[None]] | None = None
        self.loop = asyncio.get_event_loop()
        self._mqtt = mqtt
        self.iot_id = cloud_device.iot_id
        self.device = cloud_device
        self._command_futures = {}
        self._commands: MammotionCommand = MammotionCommand(
            cloud_device.device_name,
            int(mqtt.cloud_client.mammotion_http.response.data.userInformation.userAccount),
        )
        self.currentID = ""
        self._mqtt.mqtt_message_event.add_subscribers(self._parse_message_for_device)
        self._mqtt.mqtt_properties_event.add_subscribers(self._parse_message_properties_for_device)
        self._mqtt.mqtt_status_event.add_subscribers(self._parse_message_status_for_device)
        self._mqtt.mqtt_device_event.add_subscribers(self._parse_device_event_for_device)
        self._mqtt.on_ready_event.add_subscribers(self.on_ready)
        self._mqtt.on_disconnected_event.add_subscribers(self.on_disconnect)
        self._mqtt.on_connected_event.add_subscribers(self.on_connect)
        self.set_queue_callback(self.queue_command)

    def __del__(self) -> None:
        """Cleanup subscriptions."""
        self._mqtt.on_ready_event.remove_subscribers(self.on_ready)
        self._mqtt.on_disconnected_event.remove_subscribers(self.on_disconnect)
        self._mqtt.on_connected_event.remove_subscribers(self.on_connect)
        self._mqtt.mqtt_message_event.remove_subscribers(self._parse_message_for_device)
        self._mqtt.mqtt_properties_event.remove_subscribers(self._parse_message_properties_for_device)
        self._mqtt.mqtt_status_event.remove_subscribers(self._parse_message_status_for_device)
        self._mqtt.mqtt_device_event.remove_subscribers(self._parse_device_event_for_device)
        self._state_manager.cloud_queue_command_callback.remove_subscribers(self.queue_command)

    @property
    def command_sent_time(self) -> float:
        return self._mqtt.command_sent_time

    def set_notification_callback(self, func: Callable[[tuple[str, Any | None]], Awaitable[None]]) -> None:
        self._state_manager.cloud_on_notification_callback.add_subscribers(func)

    def set_queue_callback(self, func: Callable[[str, dict[str, Any]], Awaitable[None]]) -> None:
        self._state_manager.cloud_queue_command_callback.add_subscribers(func)

    async def on_ready(self) -> None:
        """Callback for when MQTT is subscribed to events."""
        if self.stopped:
            return
        try:
            if self.on_ready_callback:
                await self.on_ready_callback()
        except (DeviceOfflineException, UnretryableException):
            _LOGGER.debug("Device is offline")

    async def on_disconnect(self) -> None:
        self._mqtt.disconnect()

    async def on_connect(self) -> None:
        """On connect callback"""

    def stop(self) -> None:
        """Stop all tasks and disconnect."""
        # self._mqtt._mqtt_client.unsubscribe()
        if self.mqtt.is_connected():
            self._mqtt.disconnect()
        self.stopped = True

    async def start(self) -> None:
        """Start the device connection."""
        self.stopped = False
        if not self.mqtt.is_connected():
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, self.mqtt.connect_async)
        # else:
        #     self.mqtt._mqtt_client.thing_on_thing_enable(None)

    async def _ble_sync(self) -> None:
        pass

    async def queue_command(self, key: str, **kwargs: Any) -> None:
        # Create a future to hold the result
        _LOGGER.debug("Queueing command: %s", key)
        future = asyncio.Future()
        # Put the command in the queue as a tuple (key, command, future)
        command_bytes = getattr(self._commands, key)(**kwargs)
        await self._mqtt.command_queue.put((self.iot_id, key, command_bytes, future))
        # Wait for the future to be resolved
        try:
            await future
            return
        except asyncio.CancelledError:
            """Try again once."""
            future = asyncio.Future()
            await self._mqtt.command_queue.put((self.iot_id, key, command_bytes, future))

    def _extract_message_id(self, payload: dict) -> str:
        """Extract the message ID from the payload."""
        return payload.get("id", "")

    def _extract_encoded_message(self, payload: dict) -> str:
        """Extract the encoded message from the payload."""
        try:
            content = payload.get("data", {}).get("data", {}).get("params", {}).get("content", "")
            return str(content)
        except AttributeError:
            _LOGGER.error("Error extracting encoded message. Payload: %s", payload)
            return ""

    @staticmethod
    def dequeue_by_iot_id(queue, iot_id):
        for item in queue:
            if item.iot_id == iot_id:
                queue.remove(item)
                return item
        return None

    async def _parse_message_properties_for_device(self, event: ThingPropertiesMessage) -> None:
        if event.params.iot_id != self.iot_id:
            return
        await self.state_manager.properties(event)

    async def _parse_message_status_for_device(self, status: ThingStatusMessage) -> None:
        if status.params.iot_id != self.iot_id:
            return
        await self.state_manager.status(status)

    async def _parse_device_event_for_device(self, status: ThingStatusMessage) -> None:
        """Process device event if it matches the device's IoT ID."""
        if status.params.iot_id != self.iot_id:
            return
        await self.state_manager.device_event(status)

    async def _parse_message_for_device(self, event: ThingEventMessage) -> None:
        """Parses a message received from a device and updates internal state.

        This function processes an incoming `ThingEventMessage`, checks if the message
        is intended for this device, decodes the binary data, and updates raw data. It
        then attempts to parse the binary data into a `LubaMsg`. If parsing fails, it
        logs the exception. The function also handles setting the device product key if
        not already set and processes specific sub-messages based on their types.

        Args:
            event (ThingEventMessage): The event message received from the device.

        """
        params = event.params
        new_msg = LubaMsg()
        if event.params.iot_id != self.iot_id:
            return
        binary_data = base64.b64decode(params.value.content)
        try:
            self._update_raw_data(binary_data)
            new_msg = LubaMsg().parse(binary_data)
        except (KeyError, ValueError, IndexError, UnicodeDecodeError):
            _LOGGER.exception("Error parsing message %s", binary_data)

        if (
            self._commands.get_device_product_key() == ""
            and self._commands.get_device_name() == event.params.device_name
        ):
            self._commands.set_device_product_key(event.params.product_key)

        res = betterproto2.which_one_of(new_msg, "LubaSubMsg")
        if res[0] == "net":
            if new_msg.net.todev_ble_sync != 0 or new_msg.net.toapp_wifi_iot_status is not None:
                return

        await self._state_manager.notification(new_msg)

    @property
    def mqtt(self):
        return self._mqtt
